# -*- coding: utf-8 -*-
# Generated by Django 1.9.1 on 2016-12-04 23:13
from __future__ import unicode_literals

import re
import warnings
from html.parser import HTMLParser
from django.db import migrations
from cms.api import add_plugin
from cms.models.pluginmodel import CMSPlugin
from djangocms_text_ckeditor.cms_plugins import TextPlugin
from cmsplugin_cascade.models import CascadeElement


def _replace_text_body(old_body, input_pattern, output_tag, id_format):
    regex = re.compile(input_pattern)

    def _do_replace(match):
        before_id, old_plugin_id, after_id = match.groups()

        if not old_plugin_id:
            return ''

        bits = []

        if before_id:
            bits.append(before_id.strip())

        bits.append(id_format.format(old_plugin_id))

        if after_id:
            bits.append(after_id.strip())

        # By using .join() we ensure the correct
        # amount of spaces are used to separate the different
        # attributes.
        tag_attrs = ' '.join(bits)
        return output_tag.format(tag_attrs)

    new_body, count = regex.subn(_do_replace, old_body)
    return new_body, count


def forwards(apps, schema_editor):
    html_parser = HTMLParser()

    for cascade_element in CascadeElement.objects.all():
        if cascade_element.plugin_type != 'CarouselSlidePlugin':
            continue

        caption = cascade_element.glossary.get('caption')
        if not caption:
            continue

        text_element = add_plugin(cascade_element.placeholder, TextPlugin, cascade_element.language,
                                  target=cascade_element)

        old_body = html_parser.unescape(caption)
        new_body, count = _replace_text_body(
            old_body,
            input_pattern=r'<img ([^>]*)\bid="plugin_obj_(?P<pk>\d+)"([^>]*)/?>',
            output_tag='<cms-plugin {}></cms-plugin>',
            id_format='id="{}"',
        )
        text_element.body = new_body
        text_element.save()

        # TODO: need to be re-tested
        if False and count > 0:
            for link_element in CMSPlugin.objects.filter(parent_id__in=(cascade_element.id, cascade_element.parent_id), plugin_type='TextLinkPlugin'):
                # print("Move Link {} from {} -> {}".format(link_element.id, link_element.parent_id, text_element.id))
                link_element.move(text_element, pos='last-child')
                link_element.save()


def backwards(apps, schema_editor):
    warnings.warn("Backward migration for Carousel Slide Plugins is not supported. Please check them manually.")


class Migration(migrations.Migration):

    dependencies = [
        ('cmsplugin_cascade', '0014_glossary_field'),
    ]

    operations = [
        migrations.RunPython(forwards, reverse_code=backwards),
    ]




class PluginExtraFieldsConfig:
    """
    Each Cascade Plugin can be configured to accept extra fields, such as an ID tag, one or more
    CSS classes or inlines styles. It is possible to configure these fields globally using an
    instance of this class, or to configure them on a per site base using the appropriate
    admin's backend interface at:
    ```
    *Start › django CMS Cascade › Custom CSS classes and styles › PluginExtraFields*
    ```
    :param allow_id_tag:
        If ``True``, allows to set the ``id`` attribute in HTML elements.

    :param css_classes:
        A dictionary containing:
          ``class_names`` a comma separated string of allowed class names.
          ``multiple`` a Boolean indicating if more multiple classes are allowed concurrently.

    :param inline_styles:
        A dictionary containing:

    :param allow_override:
        If ``True``, allows to override this configuration using the admin's backend interface.
    """
    def __init__(self, allow_id_tag=False, css_classes=None, inline_styles=None, allow_override=True):
        self.allow_id_tag = allow_id_tag
        self.css_classes = dict(multiple='', class_names='') if css_classes is None else dict(css_classes)
        self.inline_styles = {} if inline_styles is None else dict(inline_styles)
        self.allow_override = allow_override



from cmsplugin_cascade.models import InlineCascadeElement, SortableInlineCascadeElement


class CascadePluginMixin:
    """
    This is the common mixin class used for both, the :class:`cmsplugin_cascade.plugin_base.CascadePluginBase`
    and the :class:`cmsplugin_cascade.strides.StridePluginBase`.
    """

    @classmethod
    def get_tag_type(self, instance):
        """
        Return the tag_type used to render this plugin.
        """
        return instance.glossary.get('tag_type', getattr(self, 'tag_type', 'div'))

    @classmethod
    def get_css_classes(cls, instance):
        """
        Returns a list of CSS classes to be added as class="..." to the current HTML tag.
        """
        css_classes = []
        if hasattr(cls, 'default_css_class'):
            css_classes.append(cls.default_css_class)
        default_css_attributes = getattr(cls, 'default_css_attributes', [])
        for attr in default_css_attributes:
            css_class = instance.glossary.get(attr)
            if isinstance(css_class, str):
                css_classes.append(css_class)
            elif isinstance(css_class, list):
                css_classes.extend(css_class)
        return css_classes

    @classmethod
    def get_inline_styles(cls, instance):
        """
        Returns a dictionary of CSS attributes to be added as style="..." to the current HTML tag.
        """
        inline_styles = getattr(cls, 'default_inline_styles', {})
        css_style = instance.glossary.get('inline_styles')
        if css_style:
            inline_styles.update(css_style)
        return inline_styles

    @classmethod
    def get_html_tag_attributes(cls, instance):
        """
        Returns a dictionary of attributes, which shall be added to the current HTML tag.
        This method normally is called by the models's property method ``html_tag_ attributes``,
        which enriches the HTML tag with those attributes converted to a list as
        ``attr1="val1" attr2="val2" ...``.
        """
        attributes = getattr(cls, 'html_tag_attributes', {})
        return dict((attr, instance.glossary.get(key, '')) for key, attr in attributes.items())


class WithInlineElementsMixin:
    """
    Plugins wishing to allow child elements as inlines, shall inherit from this
    mixin class, in order to override the serialize and deserialize methods.
    """
    @classmethod
    def get_data_representation(cls, instance):
        data = super().get_data_representation(instance)
        data.update(inlines=[ie.glossary for ie in instance.inline_elements.all()])
        return data

    @classmethod
    def add_inline_elements(cls, instance, inlines):
        for inline_glossary in inlines:
            InlineCascadeElement.objects.create(
                cascade_element=instance, glossary=inline_glossary)


class WithSortableInlineElementsMixin:
    """
    Plugins wishing to allow child elements as sortable inlines, shall inherit from this
    mixin class, in order to override the serialize and deserialize methods.
    """
    @classmethod
    def get_data_representation(cls, instance):
        data = super().get_data_representation(instance)
        data.update(inlines=[ie.glossary for ie in instance.sortinline_elements.all()])
        return data

    @classmethod
    def add_inline_elements(cls, instance, inlines):
        for order, inline_glossary in enumerate(inlines, 1):
            SortableInlineCascadeElement.objects.create(
                cascade_element=instance, glossary=inline_glossary, order=order)



from django.contrib import admin
from django.core.exceptions import ValidationError
from django.forms import Media
from django.forms.models import ModelForm
from django.utils.html import format_html, format_html_join
from django.utils.text import format_lazy
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _
from cmsplugin_cascade.models import IconFont
from cmsplugin_cascade.icon.utils import zipfile, unzip_archive


class UploadIconsForms(ModelForm):
    class Meta:
        fields = ['identifier', 'zip_file', 'is_default']

    @property
    def media(self):
        media = super().media
        try:
            css_url = self.instance.get_stylesheet_url()
            media += Media(css={'all': ['cascade/css/admin/iconfont.css', css_url]})
        except AttributeError:
            pass
        return media

    def clean(self):
        fontello_url = 'http://fontello.com/'
        cleaned_data = super().clean()
        if 'zip_file' in self.changed_data:
            try:
                label = cleaned_data['zip_file'].label
                zip_ref = zipfile.ZipFile(cleaned_data['zip_file'].file.file, 'r')
                cleaned_data.update(zip(['font_folder', 'config_data'], unzip_archive(label, zip_ref)))
            except Exception as exc:
                raise ValidationError(format_lazy(_("Can not unzip uploaded archive {}: {}"), label, exc))
            finally:
                zip_ref.close()
            try:
                css_prefix_text = cleaned_data['config_data']['css_prefix_text']
            except KeyError:
                raise ValidationError(format_lazy(
                    _("File does not seem to originate from <a href=\"{url}\" target=\"_blank\">{url}</a>"),
                    url=fontello_url,
                ))
            other_icon_fonts = IconFont.objects.all()
            if self.instance:
                other_icon_fonts = other_icon_fonts.exclude(id=self.instance.id)
            for icon_font in other_icon_fonts:
                if icon_font.config_data['css_prefix_text'] == css_prefix_text:
                    msg = format_lazy(_(
                        "Icon Font '{icon_font}' already uses CSS prefix '{css_prefix}'.<br/>" \
                        "Please reload the font from <a href=\"{url}\" target=\"_blank\">{url}</a> " \
                        "using another CSS prefix."),
                        icon_font=icon_font.identifier, css_prefix=css_prefix_text, url=fontello_url)
                    raise ValidationError(mark_safe(msg))
        return cleaned_data


@admin.register(IconFont)
class IconFontAdmin(admin.ModelAdmin):
    form = UploadIconsForms
    list_display = ['identifier', 'is_default', 'num_icons', 'css_prefix']
    readonly_fields = ['css_prefix']

    def save_model(self, request, obj, form, change):
        if 'font_folder' in form.cleaned_data and 'config_data' in form.cleaned_data:
            obj.font_folder = form.cleaned_data['font_folder']
            obj.config_data = form.cleaned_data['config_data']
        # at least one icon font must be set as default
        if self.model.objects.filter(is_default=True).count() < 1:
            obj.is_default = True
        super().save_model(request, obj, form, change)
        if obj.is_default is True and 'is_default' in form.changed_data:
            # maximum one icon font can be set as default
            self.model.objects.exclude(id=obj.id).update(is_default=False)

    def get_readonly_fields(self, request, obj=None):
        readonly_fields = list(super().get_readonly_fields(request, obj=obj))
        if obj:
            readonly_fields.append('preview_icons')
        return readonly_fields

    def preview_icons(self, obj):
        families = obj.get_icon_families()
        format_string = '<li title="{{0}}"><i class="{css_prefix_text}{{0}}"></i></li>'.format(**obj.config_data)
        return format_html('<div class="preview-iconfont">{}</div>',
            format_html_join('\n', '<h2>{}</h2><ul>{}</ul>',
                 ((src.title(), format_html_join('', format_string, ((g,) for g in glyphs)))
                 for src, glyphs in families.items())))
    preview_icons.short_description = _("Preview Icons")

    def num_icons(self, obj):
        try:
            return len(obj.config_data['glyphs'])
        except (KeyError, TypeError):
            return '–'
    num_icons.short_description = _("Number of Icons")

    def css_prefix(self, obj):
        try:
            return obj.config_data['css_prefix_text']
        except (KeyError, TypeError):
            return '–'
    css_prefix.short_description = _("CSS prefix")



# -*- coding: utf-8 -*-
# Generated by Django 1.9.5 on 2016-04-04 18:22
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import jsonfield.fields


class Migration(migrations.Migration):

    dependencies = [
        ('cmsplugin_cascade', '0007_add_proxy_models'),
    ]

    operations = [
        migrations.CreateModel(
            name='SortableInlineCascadeElement',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('glossary', jsonfield.fields.JSONField(blank=True, default={})),
                ('order', models.PositiveIntegerField(db_index=True, verbose_name='Sort by')),
                ('cascade_element', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='sortinline_elements', to='cmsplugin_cascade.CascadeElement')),
            ],
            options={
                'ordering': ('order',),
                'db_table': 'cmsplugin_cascade_sortinline',
            },
        ),
    ]



from django.db import models
from django.utils.html import mark_safe, format_html_join
from django.utils.functional import cached_property
import json
from jsonfield.fields import JSONField
from cms.models import CMSPlugin
from cms.plugin_pool import plugin_pool
from cms.utils.placeholder import get_placeholder_conf


class CascadeModelBase(CMSPlugin):
    """
    The container to hold additional HTML element tags.
    """
    class Meta:
        abstract = True

    cmsplugin_ptr = models.OneToOneField(
        CMSPlugin,
        related_name='+',
        on_delete=models.CASCADE,
        parent_link=True,
    )

    glossary = JSONField(blank=True, default={})

    def __str__(self):
        return self.plugin_class.get_identifier(self)

    @cached_property
    def plugin_class(self):
        return self.get_plugin_class()

    @property
    def tag_type(self):
        return self.plugin_class.get_tag_type(self)

    @property
    def css_classes(self):
        css_classes = self.plugin_class.get_css_classes(self)
        return mark_safe(' '.join(c for c in css_classes if c))

    @property
    def inline_styles(self):
        inline_styles = self.plugin_class.get_inline_styles(self)
        return format_html_join(' ', '{0}: {1};', (s for s in inline_styles.items() if s[1]))

    @property
    def html_tag_attributes(self):
        attributes = self.plugin_class.get_html_tag_attributes(self)
        joined = format_html_join(' ', '{0}="{1}"', ((attr, val) for attr, val in attributes.items() if val))
        if joined:
            return mark_safe(' ' + joined)
        return ''

    def get_parent_instance(self):
        for model in CascadeModelBase._get_cascade_elements():
            try:
                return model.objects.get(id=self.parent_id)
            except model.DoesNotExist:
                continue
        # in case our plugin is the child of a TextPlugin, return its grandparent
        parent = self.get_parent()
        if parent and parent.plugin_type == 'TextPlugin':
            grandparent_id = self.get_parent().parent_id
            for model in CascadeModelBase._get_cascade_elements():
                try:
                    return model.objects.get(id=grandparent_id)
                except model.DoesNotExist:
                    continue

    def get_parent_glossary(self):
        """
        Return the glossary from the parent of this object. If there is no parent, retrieve
        the glossary from the placeholder settings, if configured.
        """
        parent = self.get_parent_instance()
        if parent:
            return parent.get_complete_glossary()
        # otherwise use self.placeholder.glossary as the starting dictionary
        template = self.placeholder.page.template if self.placeholder.page else None
        return get_placeholder_conf('glossary', self.placeholder.slot, template=template, default={})

    def get_complete_glossary(self):
        """
        Return the parent glossary for this model object merged with the current object.
        This is done by starting from the root element down to the current element and enriching
        the glossary with each models's own glossary.
        """
        if not hasattr(self, '_complete_glossary_cache'):
            self._complete_glossary_cache = self.get_parent_glossary().copy()
            self._complete_glossary_cache.update(self.glossary or {})
        return self._complete_glossary_cache

    def get_num_children(self):
        """
        Returns the number of children for this plugin instance.
        """
        return self.get_children().count()

    def sanitize_children(self):
        """
        Recursively walk down the plugin tree and invoke method ``save(sanitize_only=True)`` for
        each child.
        """
        for model in CascadeModelBase._get_cascade_elements():
            # execute query to not iterate over SELECT ... FROM while updating other models
            children = list(model.objects.filter(parent_id=self.id))
            for child in children:
                child.save(sanitize_only=True)
                child.sanitize_children()
                
    @classmethod
    def from_db(cls, db, field_names, values):
        instance = cls(*values)
        if isinstance(instance.glossary, str):
           instance.glossary = json.loads(instance.glossary)
        return instance
                
    def save(self, sanitize_only=False, *args, **kwargs):
        """
        A hook which let the plugin instance sanitize the current object model while saving it.
        With ``sanitize_only=True``, the current model object only is saved when the method
        ``sanitize_model()`` from the corresponding plugin actually changed the glossary.
        """
        sanitized = self.plugin_class.sanitize_model(self)
        if sanitize_only:
            if sanitized:
                super().save(no_signals=True)
        else:
            super().save(*args, **kwargs)

    @classmethod
    def _get_cascade_elements(cls):
        """
        Returns a set of models which are derived from ``CascadeModelBase``. This set shall be used
        for traversing the plugin tree of interconnected Cascade models. Currently, Cascade itself
        offers only one model, namely ``CascadeElement``, but a third party library may extend
        ``CascadeModelBase`` and add arbitrary model fields.
        """
        if not hasattr(cls, '_cached_cascade_elements'):
            cce = set([p.model._meta.concrete_model for p in plugin_pool.get_all_plugins()
                       if issubclass(p.model, cls)])
            cls._cached_cascade_elements = cce
        return cls._cached_cascade_elements



# -*- coding: utf-8 -*-
# Generated by Django 1.11.20 on 2019-04-03 19:54
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('cmsplugin_cascade', '0022_auto_20181202_1055'),
    ]

    operations = [
        migrations.AddField(
            model_name='iconfont',
            name='is_default',
            field=models.BooleanField(default=False, help_text='Use this font as default, unless an icon font is set for the current page.', verbose_name='Default Font'),
        ),
    ]



from django.core.exceptions import ObjectDoesNotExist, ValidationError
from django.contrib.admin.sites import site as admin_site
from django.db.models.fields.related import ManyToOneRel
from django.forms import fields, Media, ModelChoiceField
from django.forms.widgets import RadioSelect
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _
from django_select2.forms import HeavySelect2Widget

from cms.utils import get_current_site
from cms.models import Page
from entangled.forms import EntangledModelFormMixin
from entangled.utils import get_related_object
from filer.models.filemodels import File as FilerFileModel
from filer.fields.file import AdminFileWidget, FilerFileField

try:
    from phonenumber_field.formfields import PhoneNumberField
except ImportError:
    PhoneNumberField = None


def format_page_link(title, path):
    html = format_html("{} ({})", mark_safe(title), path)
    return html


class PageSelect2Widget(HeavySelect2Widget):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault('data_view', 'admin:get_published_pagelist')
        super().__init__(*args, **kwargs)

    @property
    def media(self):
        parent_media = super().media
        # append jquery.init.js to enforce select2.js into the global 'jQuery' namespace
        js = list(parent_media._js) + ['admin/js/jquery.init.js']
        return Media(css=parent_media._css, js=js)

    def render(self, *args, **kwargs):
        # replace self.choices by an empty list in order to prevent building the whole optgroup
        try:
            page = Page.objects.get(pk=kwargs['value'])
        except (Page.DoesNotExist, ValueError, KeyError):
            self.choices = []
        else:
            self.choices = [(kwargs['value'], str(page))]
        return super().render(*args, **kwargs)


class LinkSearchField(ModelChoiceField):
    widget = PageSelect2Widget()

    def __init__(self, *args, **kwargs):
        queryset = Page.objects.public()
        try:
            queryset = queryset.published().on_site(get_current_site())
        except:
            choices = []  # can happen if database is not ready yet
        else:
            # set a minimal set of choices, otherwise django-select2 builds them for every published page
            choices = [(index, str(page)) for index, page in enumerate(queryset[:15])]
        kwargs.setdefault('queryset', queryset.distinct())
        super().__init__(*args, **kwargs)
        self.choices = choices


class SectionChoiceField(fields.ChoiceField):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault('choices', [('', _("Page Root"))])
        super().__init__(*args, **kwargs)

    def valid_value(self, value):
        """
        The optgroup is adjusted dynamically accroding to the selected cms_page, so always returns True
        and let `LinkForm` validate this value.
        """
        return True


class LinkForm(EntangledModelFormMixin):
    LINK_TYPE_CHOICES = [
        ('cmspage', _("CMS Page")),
        ('download', _("Download File")),
        ('exturl', _("External URL")),
        ('email', _("Mail To")),
    ]
    if PhoneNumberField:
        LINK_TYPE_CHOICES.append(('phonenumber', _("Phone number")))

    link_type = fields.ChoiceField(
        label=_("Link"),
        help_text=_("Type of link"),
    )

    cms_page = LinkSearchField(
        required=False,
        label='',
        help_text=_("An internal link onto any CMS page of this site"),
    )

    section = SectionChoiceField(
        required=False,
        label='',
        help_text=_("Page bookmark"),
    )

    download_file = ModelChoiceField(
        label='',
        queryset=FilerFileModel.objects.all(),
        widget=AdminFileWidget(ManyToOneRel(FilerFileField, FilerFileModel, 'id'), admin_site),
        required=False,
        help_text=_("An internal link onto a file from filer"),
    )

    ext_url = fields.URLField(
        required=False,
        label=_("URL"),
        help_text=_("Link onto external page"),
    )

    mail_to = fields.EmailField(
        required=False,
        label=_("Email"),
        help_text=_("Open Email program with this address"),
    )

    if PhoneNumberField:
        phone_number = PhoneNumberField(
            required=False,
            label=_("Phone Number"),
            help_text=_("International phone number, ex. +1 212 555 2368."),
        )

    link_target = fields.ChoiceField(
        choices=[
            ('', _("Same Window")),
            ('_blank', _("New Window")),
            ('_parent', _("Parent Window")),
            ('_top', _("Topmost Frame")),
        ],
        label=_("Link Target"),
        widget=RadioSelect,
        required=False,
        help_text=_("Open Link in other target."),
    )

    link_title = fields.CharField(
        label=_("Title"),
        required=False,
        help_text=_("Link's Title"),
    )

    class Meta:
        entangled_fields = {'glossary': ['link_type', 'cms_page', 'section', 'download_file', 'ext_url', 'mail_to',
                                         'link_target', 'link_title']}
        if PhoneNumberField:
            entangled_fields['glossary'].append('phone_number')

    def __init__(self, *args, **kwargs):
        link_type_choices = []
        if not getattr(self, 'require_link', True):
            link_type_choices.append(('', _("No Link")))
            self.declared_fields['link_type'].required = False
        link_type_choices.extend(self.LINK_TYPE_CHOICES)
        self.declared_fields['link_type'].choices = link_type_choices
        self.declared_fields['link_type'].initial = link_type_choices[0][0]
        instance = kwargs.get('instance')
        if instance and instance.glossary.get('link_type') == 'cmspage':
            self._preset_section(instance)
        super().__init__(*args, **kwargs)

    def _preset_section(self, instance):
        """
        Field ``cms_page`` may refer onto any CMS page, which itself may contain bookmarks. This method
        creates the list of bookmarks.
        """
        self.base_fields['section'].choices = self.base_fields['section'].choices[:1]
        try:
            cascade_page = get_related_object(instance.glossary, 'cms_page').cascadepage
            for key, val in cascade_page.glossary.get('element_ids', {}).items():
                self.base_fields['section'].choices.append((key, val))
        except (AttributeError, ObjectDoesNotExist):
            pass

    def _post_clean(self):
        super()._post_clean()
        error = None
        empty_fields = [None, '']
        link_type = self.cleaned_data['glossary'].get('link_type')
        if link_type == 'cmspage':
            if self.cleaned_data['glossary'].get('cms_page', False) in empty_fields:
                error = ValidationError(_("CMS page to link to is missing."), code='required')
                self.add_error('cms_page', error)
        elif link_type == 'download':
            if self.cleaned_data['glossary'].get('download_file', False) in empty_fields:
                error = ValidationError(_("File for download is missing."), code='required')
                self.add_error('download_file', error)
        elif link_type == 'exturl':
            ext_url = self.cleaned_data['glossary'].get('ext_url', False)
            if ext_url in empty_fields:
                error = ValidationError(_("No valid URL provided."), code='required')
                self.add_error('ext_url', error)
        elif link_type == 'email':
            if self.cleaned_data['glossary'].get('mail_to', False) in empty_fields:
                error = ValidationError(_("No email address provided."), code='required')
                self.add_error('mail_to', error)
        elif link_type == 'phonenumber':
            if self.cleaned_data['glossary'].get('phone_number', False) in empty_fields:
                error = ValidationError(_("No phone number provided."), code='required')
                self.add_error('phone_number', error)

    def clean_phone_number(self):
        return str(self.cleaned_data['phone_number'])

    @classmethod
    def unset_required_for(cls, sharable_fields):
        """
        Fields borrowed by `SharedGlossaryAdmin` to build its temporary change form, only are
        required if they are declared in `sharable_fields`. Otherwise just deactivate them.
        """
        if 'link_content' in cls.base_fields and 'link_content' not in sharable_fields:
            cls.base_fields['link_content'].required = False
        if 'link_type' in cls.base_fields and 'link' not in sharable_fields:
            cls.base_fields['link_type'].required = False


class TextLinkFormMixin(EntangledModelFormMixin):
    link_content = fields.CharField(
        label=_("Link Content"),
        widget=fields.TextInput(attrs={'id': 'id_name'}),  # replace auto-generated id so that CKEditor automatically transfers the text into this input field
        help_text=_("Content of Link"),
    )

    class Meta:
        entangled_fields = {'glossary': ['link_content']}




class AppSettings:

    def _setting(self, name, default=None):
        from django.conf import settings
        return getattr(settings, name, default)

    @property
    def CASCADE_PLUGINS(self):
        return self._setting('CMSPLUGIN_CASCADE_PLUGINS', (
            'cmsplugin_cascade.generic',
            'cmsplugin_cascade.icon',
            'cmsplugin_cascade.link',
        ))

    @property
    def CMSPLUGIN_CASCADE(self):
        import os
        from collections import OrderedDict
        from importlib import import_module
        from django.forms.fields import NumberInput
        from django.core.exceptions import ImproperlyConfigured
        from django.utils.translation import gettext_lazy
        from cmsplugin_cascade.fields import (ColorField, SelectTextAlignField, SelectOverflowField, SizeField,
                                              BorderChoiceField)

        if hasattr(self, '_config_CMSPLUGIN_CASCADE'):
            return self._config_CMSPLUGIN_CASCADE

        INSTALLED_APPS = self._setting('INSTALLED_APPS')
        config = self._setting('CMSPLUGIN_CASCADE', {})
        config.setdefault('alien_plugins', ['TextPlugin'])
        config.setdefault('color_picker_with_alpha', False)
        config.setdefault('plugin_prefix', None)

        config.setdefault('plugins_with_extra_fields', {})
        if 'cmsplugin_cascade.extra_fields' in INSTALLED_APPS:
            from cmsplugin_cascade.extra_fields.config import PluginExtraFieldsConfig

            plugins_with_extra_fields = config['plugins_with_extra_fields']
            plugins_with_extra_fields.setdefault('SimpleWrapperPlugin', PluginExtraFieldsConfig())
            for plugin, plugin_config in plugins_with_extra_fields.items():
                if not isinstance(plugin_config, PluginExtraFieldsConfig):
                    msg = "CMSPLUGIN_CASCADE['plugins_with_extra_fields']['{}'] must instantiate a class of type PluginExtraFieldsConfig"
                    raise ImproperlyConfigured(msg.format(plugin))

        config.setdefault('plugins_with_extra_mixins', {})

        config.setdefault('plugins_with_sharables', {})
        if 'cmsplugin_cascade.sharable' in INSTALLED_APPS:
            config['plugins_with_sharables'].setdefault(
                'FramedIconPlugin',
                ['font_size', 'color', 'background_color', 'text_align', 'border', 'border_radius'])

        config['exclude_hiding_plugin'] = list(config.get('exclude_hiding_plugin', []))
        config['exclude_hiding_plugin'].append('SegmentPlugin')

        config.setdefault('link_plugin_classes', (
            'cmsplugin_cascade.link.plugin_base.DefaultLinkPluginBase',
            'cmsplugin_cascade.link.forms.LinkForm'))

        config['plugins_with_bookmark'] = list(config.get('plugins_with_bookmark', []))
        config['plugins_with_bookmark'].extend(['SimpleWrapperPlugin', 'HeadingPlugin'])
        config.setdefault('bookmark_prefix', '')

        config.setdefault('extra_inline_styles', OrderedDict())
        extra_inline_styles = config['extra_inline_styles']
        extra_inline_styles.setdefault(
            'Margins',
            (['margin-top', 'margin-right', 'margin-bottom', 'margin-left'], SizeField))
        extra_inline_styles.setdefault(
            'Paddings',
            (['padding-top', 'padding-right', 'padding-bottom', 'padding-left'], SizeField))
        extra_inline_styles.setdefault(
            'Widths',
            (['min-width', 'width', 'max-width'], SizeField))
        extra_inline_styles.setdefault(
            'Heights',
            (['min-height', 'height', 'max-height'], SizeField))
        extra_inline_styles.setdefault(
            'Text Alignement',
            (['text-align'], SelectTextAlignField))
        extra_inline_styles.setdefault(
            'Font Size',
            (['font-size'], SizeField))
        extra_inline_styles.setdefault(
            'Line Height',
            (['line-height'], NumberInput))
        extra_inline_styles.setdefault(
            'Colors',
            (['color', 'background-color'], ColorField))
        extra_inline_styles.setdefault(
            'Border',
            (['border', 'border-top', 'border-right', 'border-bottom', 'border-left'], BorderChoiceField))
        extra_inline_styles.setdefault(
            'Border Radius',
            (['border-radius'], SizeField))
        extra_inline_styles.setdefault(
            'Overflow',
            (['overflow', 'overflow-x', 'overflow-y'], SelectOverflowField))

        if 'cmsplugin_cascade.segmentation' in INSTALLED_APPS:
            config.setdefault('segmentation_mixins', [
                ('cmsplugin_cascade.segmentation.mixins.EmulateUserModelMixin',
                 'cmsplugin_cascade.segmentation.mixins.EmulateUserAdminMixin')])

        config.setdefault(
            'icon_font_root',
            os.path.abspath(os.path.join(self._setting('MEDIA_ROOT'), 'icon_fonts')))

        config.setdefault('plugins_with_extra_render_templates', {})
        config['plugins_with_extra_render_templates'].setdefault(
            'TextLinkPlugin',
            [('cascade/link/text-link.html', gettext_lazy("default")),
             ('cascade/link/text-link-linebreak.html', gettext_lazy("with line break")),])
        config['plugins_with_extra_render_templates'].setdefault(
            'LeafletPlugin',
            [('cascade/plugins/leaflet.html', gettext_lazy("default")),
             ('cascade/plugins/googlemap.html', gettext_lazy("Google Map")),])

        config.setdefault('allow_plugin_hiding', False)

        config.setdefault('cache_strides', True)

        config.setdefault('register_page_editor', True)

        for module_name in self.CASCADE_PLUGINS:
            try:
                settings_module = import_module('{}.settings'.format(module_name))
                getattr(settings_module, 'set_defaults')(config)
            except (ImportError, AttributeError):
                continue

        self._config_CMSPLUGIN_CASCADE = config
        return config

    @property
    def CSS_PREFIXES(self):
        return {
            'image_set': ['-webkit-image-set', '-moz-image-set', '-o-image-set', '-ms-image-set', 'image-set'],
        }

    @property
    def RESPONSIVE_IMAGE_MAX_STEPS(self):
        """
        Responsive images are offered in a set of various widths. This number specifies the maximum number of
        generated thumbnails for a specific ``srcset`` of an image.
        """
        return 12

    @property
    def RESPONSIVE_IMAGE_STEP_SIZE(self):
        """
        Responsive images are offered in a set of various widths. This number specifies the minimum step width
        in pixels between the generated thumbnails for a specific ``srcset`` of an image. If the resulting number
        of steps would exceed ``RESPONSIVE_IMAGE_MAX_STEPS``, then a higher step width is used.
        """
        return 50

import sys  # noqa
app_settings = AppSettings()
app_settings.__name__ = __name__
sys.modules[__name__] = app_settings



from django.apps import AppConfig, apps
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.db.models.signals import pre_migrate, post_migrate
from django.db.utils import DatabaseError
from django.urls import reverse
from django.utils.translation import gettext_lazy as _


class CascadeConfig(AppConfig):
    name = 'cmsplugin_cascade'
    verbose_name = _("django CMS Cascade")
    default_permissions = ('add', 'change', 'delete')

    def ready(self):
        if 'django_select2' not in settings.INSTALLED_APPS:
            raise ImproperlyConfigured('django_select2 not configured')

        stylesSet = str(settings.CKEDITOR_SETTINGS.get('stylesSet'))
        if stylesSet != 'default:{}'.format(reverse('admin:cascade_texteditor_config')):
            msg = "settings.CKEDITOR_SETTINGS['stylesSet'] should be `format_lazy('default:{}', reverse_lazy('admin:cascade_texteditor_config'))`"
            raise ImproperlyConfigured(msg)

        pre_migrate.connect(self.__class__.pre_migrate, sender=self)
        post_migrate.connect(self.__class__.post_migrate, sender=self)

    @classmethod
    def pre_migrate(cls, sender=None, **kwargs):
        """
        Iterate over contenttypes and remove those not in proxy models
        """
        ContentType = apps.get_model('contenttypes', 'ContentType')
        try:
            queryset = ContentType.objects.filter(app_label=sender.label)
            for ctype in queryset.exclude(model__in=sender.get_proxy_models().keys()):
                model = ctype.model_class()
                if model is None:
                    sender.revoke_permissions(ctype)
                    ContentType.objects.get(app_label=sender.label, model=ctype).delete()
        except DatabaseError:
            return

    @classmethod
    def post_migrate(cls, sender=None, **kwargs):
        """
        Iterate over fake_proxy_models and add contenttypes and permissions for missing proxy
        models, if this has not been done by Django yet
        """
        ContentType = apps.get_model('contenttypes', 'ContentType')

        for model_name, proxy_model in sender.get_proxy_models().items():
            ctype, created = ContentType.objects.get_or_create(app_label=sender.label, model=model_name)
            if created:
                sender.grant_permissions(proxy_model)

    def get_proxy_models(self):
        from cmsplugin_cascade.plugin_base import fake_proxy_models

        proxy_models = {}
        for model_name in fake_proxy_models.keys():
            proxy_model = self.get_model(model_name)
            model_name = proxy_model._meta.model_name  # the model_name in lowercase normally
            proxy_models[model_name] = proxy_model
        return proxy_models

    def grant_permissions(self, proxy_model):
        """
        Create the default permissions for the just added proxy model
        """
        ContentType = apps.get_model('contenttypes', 'ContentType')
        try:
            Permission = apps.get_model('auth', 'Permission')
        except LookupError:
            return

        # searched_perms will hold the permissions we're looking for as (content_type, (codename, name))
        searched_perms = []
        ctype = ContentType.objects.get_for_model(proxy_model)
        for perm in self.default_permissions:
            searched_perms.append((
                '{0}_{1}'.format(perm, proxy_model._meta.model_name),
                "Can {0} {1}".format(perm, proxy_model._meta.verbose_name_raw)
            ))

        all_perms = set(Permission.objects.filter(
            content_type=ctype,
        ).values_list(
            'content_type', 'codename'
        ))
        permissions = [
            Permission(codename=codename, name=name, content_type=ctype)
            for codename, name in searched_perms if (ctype.pk, codename) not in all_perms
        ]
        Permission.objects.bulk_create(permissions)

    def revoke_permissions(self, ctype):
        """
        Remove all permissions for the content type to be removed
        """
        ContentType = apps.get_model('contenttypes', 'ContentType')
        try:
            Permission = apps.get_model('auth', 'Permission')
        except LookupError:
            return

        codenames = ['{0}_{1}'.format(perm, ctype) for perm in self.default_permissions]
        cascade_element = apps.get_model(self.label, 'cascadeelement')
        element_ctype = ContentType.objects.get_for_model(cascade_element)
        Permission.objects.filter(content_type=element_ctype, codename__in=codenames).delete()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import jsonfield.fields


class Migration(migrations.Migration):

    dependencies = [
        ('cmsplugin_cascade', '0002_auto_20150530_1018'),
    ]

    operations = [
        migrations.CreateModel(
            name='InlineCascadeElement',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('glossary', jsonfield.fields.JSONField(default={}, blank=True)),
                ('cascade_element', models.ForeignKey(related_name='inline_elements', to='cmsplugin_cascade.CascadeElement', on_delete=models.CASCADE)),
            ],
            options={
                'db_table': 'cmsplugin_cascade_inline',
            },
            bases=(models.Model,),
        ),
    ]



import io
import json
import os
from django.conf import settings
from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _
from cms.menu_bases import CMSAttachMenu
from menus.base import NavigationNode
from menus.menu_pool import menu_pool


class DocumentationMenu(CMSAttachMenu):
    name = _("Documentation Menu")  # give the menu a name this is required.

    def get_nodes(self, request, root_page):
        """
        This method is used to build the menu tree.
        """
        nodes = []
        docsmap_file = os.path.join(settings.SPHINX_DOCS_ROOT, 'docsmap.json')
        if not os.path.exists(docsmap_file):
            return nodes
        with io.open(docsmap_file) as fh:
            docs_map = json.load(fh, encoding='utf-8')

        for counter, items in enumerate(docs_map.items(), 1):
            bits = items[0].split('/')
            if len(bits) == 1 and bits[0] == 'index' or len(bits) == 2 and bits[1] != 'index':
                continue
            node = NavigationNode(
                title=items[1],
                url=reverse_lazy('sphinx-documentation', args=(bits[0],)),
                id=counter,
            )
            nodes.append(node)
        return nodes

menu_pool.register_menu(DocumentationMenu)



from django.core.exceptions import ObjectDoesNotExist
from django.utils.functional import cached_property
from django.utils.safestring import mark_safe

from entangled.utils import get_related_object
from cmsplugin_cascade.plugin_base import CascadePluginBase
from filer.models.filemodels import File as FilerFileModel


class LinkPluginBase(CascadePluginBase):
    allow_children = False
    parent_classes = []
    require_parent = False
    ring_plugin = 'LinkPluginBase'
    raw_id_fields = ['download_file']
    html_tag_attributes = {'title': 'title', 'target': 'target'}

    class Media:
        css = {'all': ['cascade/css/admin/linkplugin.css']}
        js = ['admin/js/jquery.init.js', 'cascade/js/admin/linkplugin.js']

    @classmethod
    def get_link(cls, obj):
        linktype = obj.glossary.get('link_type')
        if linktype == 'exturl':
            return '{ext_url}'.format(**obj.glossary)
        if linktype == 'email':
            return 'mailto:{mail_to}'.format(**obj.glossary)
        if linktype == 'phonenumber':
            return 'tel:{phone_number}'.format(**obj.glossary)

        # otherwise resolve by model
        if linktype == 'cmspage':
            relobj = get_related_object(obj.glossary, 'cms_page')
            if relobj:
                href = relobj.get_absolute_url()
                try:
                    element_ids = relobj.cascadepage.glossary['element_ids']
                    section = obj.glossary['section']
                    href = '{}#{}'.format(href, element_ids[section])
                except (KeyError, ObjectDoesNotExist):
                    pass
                return href
        elif linktype == 'download':
            relobj = get_related_object(obj.glossary, 'download_file')
            if isinstance(relobj, FilerFileModel):
                return relobj.url
        return linktype


class DefaultLinkPluginBase(LinkPluginBase):
    """
    The default `LinkPluginBase` class. It is injected by the class creator in link.config
    """
    ring_plugin = 'LinkPluginBase'


class LinkElementMixin:
    """
    A mixin class to convert a CascadeElement into a proxy model for rendering the ``<a>`` element.
    Note that a Link inside the Text Editor Plugin is rendered using ``str(instance)`` rather
    than ``instance.content``.
    """
    def __str__(self):
        return self.plugin_class.get_identifier(self)

    @property
    def link(self):
        return self.plugin_class.get_link(self)

    @property
    def content(self):
        return mark_safe(self.glossary.get('link_content', ''))

    @cached_property
    def download_name(self):
        link_type = self.glossary.get('link_type')
        if link_type == 'download':
            relobj = get_related_object(self.glossary, 'download_file')
            if isinstance(relobj, FilerFileModel):
                return mark_safe(relobj.original_filename)



import json

from django.forms import widgets
from django.forms.fields import CharField, BooleanField
from django.db.models.fields.related import ManyToOneRel
from django.contrib.admin import StackedInline
from django.core.exceptions import ValidationError
from django.utils.html import strip_tags, strip_spaces_between_tags
from django.utils.safestring import mark_safe
from django.utils.translation import ngettext_lazy, gettext_lazy as _
from filer.fields.image import FilerImageField, AdminImageFormField
from filer.settings import settings as filer_settings
from filer.utils.loader import load_model
from cms.plugin_pool import plugin_pool
from djangocms_text_ckeditor.fields import HTMLFormField

from cmsplugin_cascade import app_settings
from cmsplugin_cascade.fields import HiddenDictField, SizeField, MultiSizeField
from cmsplugin_cascade.forms import CascadeModelForm, CascadeModelFormMixin
from cmsplugin_cascade.image import ImagePropertyMixin
from cmsplugin_cascade.mixins import WithInlineElementsMixin
from cmsplugin_cascade.models import InlineCascadeElement
from cmsplugin_cascade.plugin_base import CascadePluginBase, create_proxy_model
from cmsplugin_cascade.utils import compute_aspect_ratio, get_image_size, parse_responsive_length

Image = load_model(filer_settings.FILER_IMAGE_MODEL)


class MarkerModelMixin:
    @property
    def data(self):
        return mark_safe(json.dumps(self.glossary))


class MarkerForm(CascadeModelForm):
    title = CharField(
        label=_("Marker Title"),
        widget=widgets.TextInput(attrs={'size': 60}),
        help_text=_("Please choose a title, then go to the map to set a marker pin")
    )

    use_icon = BooleanField(
        label=_("Use customized marker icon"),
        initial=False,
        required=False,
    )

    marker_image = AdminImageFormField(
        ManyToOneRel(FilerImageField, Image, 'file_ptr'),
        Image.objects.all(),
        label=_("Marker Image"),
        required=False,
        to_field_name='image_file',
    )

    marker_width = SizeField(
        label=_("Marker Width"),
        allowed_units=['px'],
        required=False,
        help_text=_("Width of the marker icon in pixels."),
    )

    marker_anchor = MultiSizeField(
        ['left', 'top'],
        label=_("Marker Anchor"),
        allowed_units=['px', '%'],
        required=False,
        help_text=_("The coordinates of the icon's anchor (relative to its top left corner)."),
    )

    popup_text = HTMLFormField(
        required=False,
        help_text=_("Optional rich text to display in popup."),
    )

    position = HiddenDictField()

    class Meta:
        entangled_fields = {'glossary': ['title', 'use_icon', 'marker_image', 'marker_width', 'marker_anchor',
                                         'popup_text', 'position']}

    def clean(self):
        cleaned_data = super().clean()
        try:
            position = cleaned_data['position']
            if isinstance(position, str):
                position = json.loads(position)
            elif not isinstance(position, dict):
                raise ValueError
        except (ValueError, KeyError):
            raise ValidationError("Invalid internal position data. Check your Javascript imports.")
        else:
            if 'lat' not in position or 'lng' not in position:
                # place the marker in the center of the current map
                position = {k: v for k, v in self.instance.cascade_element.glossary['map_position'].items()
                            if k in ['lat', 'lng']}
                cleaned_data['position'] = position

        popup_text = cleaned_data.pop('popup_text', '')
        if strip_tags(popup_text):
            cleaned_data['popup_text'] = strip_spaces_between_tags(popup_text)
        return cleaned_data


class MarkerInline(StackedInline):
    model = InlineCascadeElement
    form = MarkerForm
    raw_id_fields = ['marker_image']
    verbose_name = _("Marker")
    verbose_name_plural = _("Markers")
    extra = 0


class LeafletFormMixin(CascadeModelFormMixin):
    map_width = SizeField(
        label=_("Map Width"),
        allowed_units=['px', '%'],
        initial='100%',
        help_text=_("Set the map width in percent relative to containing element."),
    )

    map_height = SizeField(
        label=_("Adapt Map Height"),
        allowed_units=['px', '%'],
        initial='400px',
        help_text=_("Set a fixed height in pixels, or percent relative to the map width."),
    )

    map_min_height = SizeField(
        label=_("Adapt Map Minimum Height"),
        allowed_units=['px'],
        required=False,
        help_text=_("Optional, set a minimum height in pixels."),
    )

    scroll_wheel_zoom = BooleanField(
        label=_("Zoom by scrolling wheel"),
        initial=True,
        required=False,
        help_text=_("Zoom into map on mouse over by scrolling wheel."),
    )

    map_position = HiddenDictField(
        initial=app_settings.CMSPLUGIN_CASCADE['leaflet']['default_position'],
    )

    class Meta:
        entangled_fields = {'glossary': ['map_width', 'map_height', 'map_position', 'map_min_height',
                                         'scroll_wheel_zoom']}

    def clean(self):
        cleaned_data = super().clean()
        try:
            if isinstance(cleaned_data['map_position'], str):
                cleaned_data['map_position'] = json.loads(cleaned_data['map_position'])
            elif not isinstance(cleaned_data['map_position'], dict):
                raise ValueError
        except (ValueError, KeyError):
            raise ValidationError("Invalid internal position data. Check your Javascript imports.")
        return cleaned_data


class LeafletModelMixin:
    @property
    def map_position(self):
        return mark_safe(json.dumps(self.glossary.get('map_position', {})))


class LeafletPlugin(WithInlineElementsMixin, CascadePluginBase):
    name = _("Map")
    parent_classes = None
    require_parent = False
    allow_children = False
    change_form_template = 'cascade/admin/leaflet_plugin_change_form.html'
    ring_plugin = 'LeafletPlugin'
    form = LeafletFormMixin
    admin_preview = False
    render_template = 'cascade/plugins/leaflet.html'
    inlines = [MarkerInline]
    model_mixins = (LeafletModelMixin,)
    settings = mark_safe(json.dumps(app_settings.CMSPLUGIN_CASCADE['leaflet']))

    class Media:
        css = {'all': [
            'node_modules/leaflet/dist/leaflet.css',
            'node_modules/leaflet-easybutton/src/easy-button.css',
            'cascade/css/admin/leafletplugin.css',
        ]}
        js = [
            'node_modules/leaflet/dist/leaflet.js',
            'node_modules/leaflet-easybutton/src/easy-button.js',
            'admin/js/jquery.init.js',
            'cascade/js/admin/leafletplugin.js',
        ]

    def add_view(self, request, form_url='', extra_context=None):
        extra_context = dict(extra_context or {}, settings=self.settings)
        return super().add_view(request, form_url, extra_context)

    def change_view(self, request, object_id, form_url='', extra_context=None):
        extra_context = dict(extra_context or {}, settings=self.settings)
        return super().change_view(request, object_id, form_url, extra_context)

    def render(self, context, instance, placeholder):
        marker_instances = []
        for inline_element in instance.inline_elements.all():
            try:
                ProxyModel = create_proxy_model('LeafletMarker',
                                                (ImagePropertyMixin, MarkerModelMixin),
                                                InlineCascadeElement,
                                                module=__name__)
                marker = ProxyModel(id=inline_element.id, glossary=inline_element.glossary)
                try:
                    aspect_ratio = compute_aspect_ratio(marker.image)
                    width = parse_responsive_length(marker.glossary.get('marker_width') or '25px')
                    marker.size = list(get_image_size(width[0], (None, None), aspect_ratio))
                    marker.size2x = 2 * marker.size[0], 2 * marker.size[1]
                except Exception:
                    # if accessing the image file fails, skip size computations
                    pass
                else:
                    try:
                        marker_anchor = marker.glossary['marker_anchor']
                        top = parse_responsive_length(marker_anchor['top'])
                        left = parse_responsive_length(marker_anchor['left'])
                        if top[0] is None or left[0] is None:
                            left = width[0] * left[1]
                            top = width[0] * aspect_ratio * top[1]
                        else:
                            left, top = left[0], top[0]
                        marker.anchor = [left, top]
                    except Exception:
                        pass
                marker_instances.append(marker)
            except (KeyError, AttributeError):
                pass

        context.update(dict(instance=instance,
                            placeholder=placeholder,
                            settings=self.settings,
                            config=app_settings.CMSPLUGIN_CASCADE['leaflet'],
                            markers=marker_instances))
        return context

    @classmethod
    def get_css_classes(cls, obj):
        css_classes = cls.super(LeafletPlugin, cls).get_css_classes(obj)
        css_class = obj.glossary.get('css_class')
        if css_class:
            css_classes.append(css_class)
        return css_classes

    @classmethod
    def get_identifier(cls, obj):
        num_elems = obj.inline_elements.count()
        content = ngettext_lazy("with {0} marker", "with {0} markers", num_elems).format(num_elems)
        return mark_safe(content)

    @classmethod
    def get_data_representation(cls, instance):
        data = super().get_data_representation(instance)
        data.update(inlines=[ie.glossary for ie in instance.inline_elements.all()])
        return data

plugin_pool.register_plugin(LeafletPlugin)



"""
See PEP 386 (https://www.python.org/dev/peps/pep-0386/)

Release logic:
 1. Remove ".devX" from __version__ (below)
 2. Remove ".devX" latest version in docs/source/changelog.rst
 3. git add cmsplugin_cascade/__init__.py docs/source/changelog.rst
 4. git commit -m 'Bump to <version>'
 5. git tag <version>
 6. git push
 7. assure that all tests pass on https://travis-ci.org/jrief/djangocms-cascade
 8. git push --tags
 9. rm -rf dist
10. python setup.py sdist
11. twine upload dist/*
12. bump the version, append ".dev0" to __version__
13. Add a new heading to docs/source/changelog.rst, named "<next-version>.dev0"
14. git add cmsplugin_cascade/__init__.py docs/source/changelog.rst
15. git commit -m 'Start with <version>'
16. git push
"""
__version__ = "1.3.7"

default_app_config = 'cmsplugin_cascade.apps.CascadeConfig'



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import jsonfield.fields
import cmsplugin_cascade.mixins
import cmsplugin_cascade.link.plugin_base
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('sites', '0001_initial'),
        ('cms', '0003_auto_20140926_2347'),
    ]

    operations = [
        migrations.CreateModel(
            name='CascadeElement',
            fields=[
                ('cmsplugin_ptr', models.OneToOneField(parent_link=True, related_name='+', primary_key=True, serialize=False, to='cms.CMSPlugin', on_delete=models.CASCADE)),
                ('glossary', jsonfield.fields.JSONField(default={}, null=True, blank=True)),
            ],
            options={
                'db_table': 'cmsplugin_cascade_element',
            },
            bases=('cms.cmsplugin',),
        ),
        migrations.CreateModel(
            name='PluginExtraFields',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('plugin_type', models.CharField(db_index=True, max_length=50, verbose_name='Plugin Name', choices=[(b'BootstrapButtonPlugin', b'Bootstrap Button'), (b'SimpleWrapperPlugin', b'Bootstrap Simple Wrapper'), (b'BootstrapRowPlugin', b'Bootstrap Row'), (b'BootstrapPicturePlugin', b'Bootstrap Picture'), (b'BootstrapContainerPlugin', b'Bootstrap Container'), (b'BootstrapColumnPlugin', b'Bootstrap Column')])),
                ('allow_id_tag', models.BooleanField(default=False)),
                ('css_classes', jsonfield.fields.JSONField(default={}, null=True, blank=True)),
                ('inline_styles', jsonfield.fields.JSONField(default={}, null=True, blank=True)),
                ('site', models.ForeignKey(verbose_name='Site', to='sites.Site', on_delete=models.CASCADE)),
            ],
            options={
                'verbose_name': 'Custom CSS classes and styles',
                'verbose_name_plural': 'Custom CSS classes and styles',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='SharableCascadeElement',
            fields=[
                ('cmsplugin_ptr', models.OneToOneField(parent_link=True, related_name='+', primary_key=True, serialize=False, to='cms.CMSPlugin', on_delete=models.CASCADE)),
                ('glossary', jsonfield.fields.JSONField(default={}, null=True, blank=True)),
            ],
            options={
                'db_table': 'cmsplugin_cascade_sharableelement',
            },
            bases=('cms.cmsplugin',),
        ),
        migrations.CreateModel(
            name='SharedGlossary',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('plugin_type', models.CharField(verbose_name='Plugin Name', max_length=50, editable=False, db_index=True)),
                ('identifier', models.CharField(unique=True, max_length=50, verbose_name='Identifier')),
                ('glossary', jsonfield.fields.JSONField(default={}, null=True, blank=True)),
            ],
            options={
                'verbose_name': 'Shared between Plugins',
                'verbose_name_plural': 'Shared between Plugins',
            },
            bases=(models.Model,),
        ),
        migrations.AlterUniqueTogether(
            name='sharedglossary',
            unique_together=set([('plugin_type', 'identifier')]),
        ),
        migrations.AddField(
            model_name='sharablecascadeelement',
            name='shared_glossary',
            field=models.ForeignKey(on_delete=django.db.models.deletion.SET_NULL, blank=True, to='cmsplugin_cascade.SharedGlossary', null=True),
            preserve_default=True,
        ),
        migrations.AlterUniqueTogether(
            name='pluginextrafields',
            unique_together=set([('plugin_type', 'site')]),
        ),
    ]



import json
import os
import shutil
from collections import OrderedDict
from urllib.parse import urljoin

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.sites.models import Site
from django.db import models
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from jsonfield.fields import JSONField
from filer.fields.file import FilerFileField
from cms.extensions import PageExtension
from cms.extensions.extension_pool import extension_pool
from cms.plugin_pool import plugin_pool
from cmsplugin_cascade.models_base import CascadeModelBase
from cmsplugin_cascade import app_settings


class SharedGlossary(models.Model):
    """
    A model class to hold glossary data shared among different plugins.
    """
    plugin_type = models.CharField(
        _("Plugin Name"),
        max_length=50,
        db_index=True,
        editable=False,
    )

    identifier = models.CharField(
        _("Identifier"),
        max_length=50,
        unique=True,
    )

    glossary = JSONField(
        null=True,
        blank=True,
        default={},
    )

    class Meta:
        unique_together = ['plugin_type', 'identifier']
        verbose_name_plural = verbose_name = _("Shared between Plugins")

    def __str__(self):
        return self.identifier

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        """
        Only entries which are declared as sharable, shall be stored in the sharable glossary.
        """
        plugin_instance = plugin_pool.get_plugin(self.plugin_type)
        glossary = dict((key, value) for key, value in self.glossary.items()
                        if key in plugin_instance.sharable_fields)
        self.glossary = glossary
        super().save(force_insert, force_update, using, update_fields)


class CascadeElement(CascadeModelBase):
    """
    The concrete model class to store arbitrary data for plugins derived from CascadePluginBase.
    """
    shared_glossary = models.ForeignKey(
        SharedGlossary,
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
    )

    class Meta:
        db_table = 'cmsplugin_cascade_element'
        verbose_name = _("Element")
        verbose_name_plural = _("Elements")

    def copy_relations(self, oldinstance):
        def init_element(inline_element):
            inline_element.pk = None
            inline_element.cascade_element = self
            inline_element.save()

        for inline_element in oldinstance.inline_elements.all():
            init_element(inline_element)
        for sortinline_element in oldinstance.sortinline_elements.all():
            init_element(sortinline_element)


class SharableCascadeElement(CascadeElement):
    """
    A proxy model which takes care of merging the glossary with its shared instance.
    """
    class Meta:
        proxy = True

    def __getattribute__(self, name):
        """
        Update glossary with content from SharedGlossary model if that exists.
        """
        attribute = object.__getattribute__(self, name)
        if name == 'glossary' and self.shared_glossary:
            attribute.update(self.shared_glossary.glossary)
        return attribute


class InlineCascadeElement(models.Model):
    cascade_element = models.ForeignKey(
        CascadeElement,
        related_name='inline_elements',
        on_delete=models.CASCADE,
    )

    glossary = JSONField(
        blank=True,
        default={},
    )

    class Meta:
        db_table = 'cmsplugin_cascade_inline'


class SortableInlineCascadeElement(models.Model):
    cascade_element = models.ForeignKey(
        CascadeElement,
        related_name='sortinline_elements',
        on_delete=models.CASCADE,
    )

    glossary = JSONField(
        blank=True,
        default={},
    )

    order = models.PositiveIntegerField(
        verbose_name=_("Sort by"),
        db_index=True,
    )

    class Meta:
        db_table = 'cmsplugin_cascade_sortinline'
        ordering = ['order']

    def __str__(self):
        return ""


class PluginExtraFields(models.Model):
    """
    Store a set of allowed extra CSS classes and inline styles to be used for Cascade plugins
    inheriting from `ExtraFieldsMixin`. Also store if individual ``id=""`` tags are allowed.
    """
    plugin_type = models.CharField(
        _("Plugin Name"),
        max_length=50,
        db_index=True,
    )

    site = models.ForeignKey(
        Site,
        verbose_name=_("Site"),
        on_delete=models.CASCADE,
    )

    allow_id_tag = models.BooleanField(default=False)

    css_classes = JSONField(
        null=True,
        blank=True,
        default={},
    )

    inline_styles = JSONField(
        null=True,
        blank=True,
        default={},
    )

    class Meta:
        verbose_name = verbose_name_plural = _("Custom CSS classes and styles")
        unique_together = ['plugin_type', 'site']

    def __str__(self):
        return str(self.name)

    @cached_property
    def name(self):
        return plugin_pool.get_plugin(self.plugin_type).name


class TextEditorConfigFields(models.Model):
    ELEMENT_CHOICES = [(c, c) for c in ['p', 'h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'pre', 'address', 'div', 'span']]

    name = models.CharField(
        _("Name"),
        max_length=50,
    )

    element_type = models.CharField(
        _("Element Type"),
        choices=ELEMENT_CHOICES,
        max_length=12,
    )

    css_classes = models.CharField(
        _("CSS classes"),
        max_length=250,
        help_text=_("Freely selectable CSS classnames for this Text-Editor Style, separated by spaces."),
    )

    class Meta:
        verbose_name = _("Text Editor Config")

    def get_config(self):
        config = {
            'name': self.name,
            'element': self.element_type,
            'attributes': {'class': self.css_classes},
        }
        return json.dumps(config)


class Segmentation(models.Model):
    class Meta:
        verbose_name = _("Segmentation")
        managed = False  # it's a dummy model
        db_table = None


class CascadeClipboard(models.Model):
    """
    A model class to persist, export and re-import the clipboard's content.
    """
    identifier = models.CharField(
        _("Identifier"),
        max_length=50,
        unique=True,
    )

    data = JSONField(
        null=True,
        blank=True,
        default={},
    )

    created_by = models.ForeignKey(
        get_user_model(),
        verbose_name=_("Created by"),
        on_delete=models.SET_NULL,
        editable=False,
        null=True,
    )

    created_at = models.DateTimeField(
        _("Created at"),
        auto_now_add=True,
        editable=False,
    )

    last_accessed_at = models.DateTimeField(
        _("Last accessed at"),
        null=True,
        default=None,
        editable=False,
    )

    class Meta:
        verbose_name = _("Persisted Clipboard Content")
        verbose_name_plural = _("Persisted Clipboard Contents")

    def __str__(self):
        return self.identifier


class FilePathField(models.FilePathField):
    """
    Implementation of `models.FilePathField` which configures the `path` argument by default
    to avoid the creation of a migration file for each change in local settings.
    """
    def __init__(self, **kwargs):
        kwargs.setdefault('path', app_settings.CMSPLUGIN_CASCADE['icon_font_root'])
        super().__init__(**kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        del kwargs['path']
        return name, path, args, kwargs


class IconFont(models.Model):
    """
    Instances of uploaded icon fonts, such as FontAwesone, MaterialIcons, etc.
    """
    identifier = models.CharField(
        _("Identifier"),
        max_length=50,
        unique=True,
        help_text=_("A unique identifier to distinguish this icon font."),
    )

    config_data = JSONField()

    zip_file = FilerFileField(
        on_delete=models.CASCADE,
        help_text=_('Upload a zip file created on <a href="http://fontello.com/" target="_blank">Fontello</a> containing fonts.')
    )

    font_folder = FilePathField(allow_files=False, allow_folders=True)

    is_default = models.BooleanField(
        _("Default Font"),
        default=False,
        help_text=_("Use this font as default, unless an icon font is set for the current page."),
    )

    class Meta:
        verbose_name = _("Uploaded Icon Font")
        verbose_name_plural = _("Uploaded Icon Fonts")

    def __str__(self):
        return self.identifier

    def get_icon_families(self):
        """
        Return an ordered dict of css classes required to render these icons
        """
        families = OrderedDict()
        for glyph in self.config_data['glyphs']:
            src = glyph.pop('src', 'default')
            families.setdefault(src, [])
            css = glyph.get('css')
            if css:
                families[src].append(css)
        return families

    def get_stylesheet_url(self):
        icon_font_url = os.path.relpath(app_settings.CMSPLUGIN_CASCADE['icon_font_root'], settings.MEDIA_ROOT)
        name = self.config_data.get('name') or 'fontello'
        parts = (icon_font_url, self.font_folder, 'css/{}.css'.format(name))
        return urljoin(settings.MEDIA_URL, '/'.join(parts))

    def config_data_as_json(self):
        data = dict(self.config_data)
        data.pop('glyphs', None)
        data['families'] = self.get_icon_families()
        return json.dumps(data)

    @classmethod
    def delete_icon_font(cls, instance=None, **kwargs):
        if isinstance(instance, cls):
            font_folder = os.path.join(app_settings.CMSPLUGIN_CASCADE['icon_font_root'], instance.font_folder)
            shutil.rmtree(font_folder, ignore_errors=True)
            try:
                temp_folder = os.path.abspath(os.path.join(font_folder, os.path.pardir))
                os.rmdir(temp_folder)
            except FileNotFoundError:
                pass

models.signals.pre_delete.connect(IconFont.delete_icon_font, dispatch_uid='delete_icon_font')


class CascadePage(PageExtension):
    """
    Keep arbitrary data tightly coupled to the CMS page.
    """
    settings = JSONField(
        blank=True,
        default={},
        help_text=_("User editable settings for this page."),
    )

    glossary = JSONField(
        blank=True,
        default={},
        help_text=_("Store for arbitrary page data."),
    )

    icon_font = models.ForeignKey(
        IconFont,
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        verbose_name=_("Icon Font"),
    )

    menu_symbol = models.CharField(
        _("Menu Symbol"),
        blank=True,
        null=True,
        max_length=32,
        help_text=_("Symbol to be used with the menu title for this page."),
    )

    class Meta:
        db_table = 'cmsplugin_cascade_page'
        verbose_name = verbose_name_plural = _("Cascade Page Settings")

    def __str__(self):
        return self.get_page().get_title()

    @classmethod
    def assure_relation(cls, cms_page):
        """
        Assure that we have a foreign key relation, pointing from CascadePage onto CMSPage.
        """
        try:
            cms_page.cascadepage
        except cls.DoesNotExist:
            cls.objects.create(extended_object=cms_page)

    @classmethod
    def delete_cascade_element(cls, instance=None, **kwargs):
        if isinstance(instance, CascadeModelBase):
            try:
                instance.placeholder.page.cascadepage.glossary['element_ids'].pop(str(instance.pk))
                instance.placeholder.page.cascadepage.save()
            except (AttributeError, KeyError):
                pass

extension_pool.register(CascadePage)
models.signals.pre_delete.connect(CascadePage.delete_cascade_element, dispatch_uid='delete_cascade_element')



from django.core.exceptions import ValidationError, ObjectDoesNotExist
from django.forms.fields import CharField
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _

from entangled.forms import EntangledModelFormMixin
from cmsplugin_cascade import app_settings
from cmsplugin_cascade.models import CascadePage


class SectionFormMixin(EntangledModelFormMixin):
    element_id = CharField(
        label=_("Element ID"),
        max_length=15,
        required=False,
        help_text=_("A unique identifier for this element.")
    )

    class Meta:
        entangled_fields = {'glossary': ['element_id']}

    def clean_element_id(self):
        element_id = self.cleaned_data['element_id']
        self.check_unique_element_id(self.instance, element_id)
        return element_id

    @classmethod
    def check_unique_element_id(cls, instance, element_id):
        """
        Check for uniqueness of the given element_id for the current page.
        Return None if instance is not yet associated with a page.
        """
        try:
            element_ids = instance.placeholder.page.cascadepage.glossary.get('element_ids', {})
        except (AttributeError, ObjectDoesNotExist):
            pass
        else:
            if element_id:
                for key, value in element_ids.items():
                    if str(key) != str(instance.pk) and element_id == value:
                        msg = _("The element ID '{}' is not unique for this page.")
                        raise ValidationError(msg.format(element_id))


class SectionModelMixin:
    def element_id(self):
        id_attr = self.glossary.get('element_id')
        if id_attr:
            return '{bookmark_prefix}{0}'.format(id_attr, **app_settings.CMSPLUGIN_CASCADE)

    def delete(self, *args, **kwargs):
        try:
            self.placeholder.page.cascadepage.glossary['element_ids'].pop(str(self.pk))
        except (AttributeError, KeyError, ObjectDoesNotExist):
            pass
        else:
            self.placeholder.page.cascadepage.save()
        super().delete(*args, **kwargs)


class SectionMixin:
    def get_form(self, request, obj=None, **kwargs):
        form = kwargs.get('form', self.form)
        assert issubclass(form, EntangledModelFormMixin), "Form must inherit from EntangledModelFormMixin"
        kwargs['form'] = type(form.__name__, (SectionFormMixin, form), {})
        return super().get_form(request, obj, **kwargs)

    @classmethod
    def get_identifier(cls, instance):
        element_id = instance.glossary.get('element_id')
        if element_id:
            return format_html('<code>id="{0}"</code>', element_id)
        return super().get_identifier(instance)

    def save_model(self, request, obj, form, change):
        super().save_model(request, obj, form, change)
        element_id = obj.glossary['element_id']
        if not change:
            # when adding a new element, `element_id` can not be validated for uniqueness
            postfix = 0
            # check if form simplewarpper has function check_unique_element_id
            if not 'check_unique_element_id' in dir(form):
                form_ = SectionFormMixin
            else:
                form_ = form
            while True:
                try:
                    form_.check_unique_element_id(obj, element_id)
                except ValidationError:
                    postfix += 1
                    element_id = '{element_id}_{0}'.format(postfix, **obj.glossary)
                else:
                    break
            if postfix:
                obj.glossary['element_id'] = element_id
                obj.save()

        cms_page = obj.placeholder.page
        if cms_page:
            # storing the element_id on a placholder only makes sense, if it is non-static
            CascadePage.assure_relation(cms_page)
            cms_page.cascadepage.glossary.setdefault('element_ids', {})
            cms_page.cascadepage.glossary['element_ids'][str(obj.pk)] = element_id
            cms_page.cascadepage.save()



from django.utils.translation import gettext_lazy as _

from cms.plugin_pool import plugin_pool
from cmsplugin_cascade.link.config import LinkPluginBase, LinkFormMixin
from cmsplugin_cascade.link.plugin_base import LinkElementMixin
from cmsplugin_cascade.icon.forms import IconFormMixin
from cmsplugin_cascade.icon.plugin_base import IconPluginMixin


class SimpleIconPlugin(IconPluginMixin, LinkPluginBase):
    name = _("Simple Icon")
    parent_classes = None
    require_parent = False
    allow_children = False
    render_template = 'cascade/plugins/simpleicon.html'
    form = type('SimpleIconForm', (LinkFormMixin, IconFormMixin), {'require_link': False})
    model_mixins = (LinkElementMixin,)
    ring_plugin = 'IconPlugin'

    class Media:
        js = ['admin/js/jquery.init.js', 'cascade/js/admin/iconplugin.js']

plugin_pool.register_plugin(SimpleIconPlugin)




from django.forms import MediaDefiningClass
from django.forms.fields import ChoiceField
from django.utils.translation import gettext_lazy as _
from django.template.loader import get_template, TemplateDoesNotExist
from entangled.forms import EntangledModelFormMixin
from cmsplugin_cascade import app_settings


class RenderTemplateFormMixin(EntangledModelFormMixin):
    render_template = ChoiceField(
        label=_("Render template"),
        help_text=_("Use alternative template for rendering this plugin."),
        required=False,
    )

    class Meta:
        entangled_fields = {'glossary': ['render_template']}


class RenderTemplateMixin(metaclass=MediaDefiningClass):
    """
    If a Cascade plugin is listed in ``settings.CMSPLUGIN_CASCADE['plugins_with_extra_render_templates']``,
    then this ``RenderTemplateMixin`` class is added automatically to its plugin class in order
    to add an additional select box used for choosing an alternative render template.
    """
    @classmethod
    def get_template_choices(cls):
        return app_settings.CMSPLUGIN_CASCADE['plugins_with_extra_render_templates'][cls.__name__]

    def get_form(self, request, obj=None, **kwargs):
        form = kwargs.get('form', self.form)
        assert issubclass(form, EntangledModelFormMixin), "Form must inherit from EntangledModelFormMixin"
        choices = self.get_template_choices()
        if isinstance(choices, (list, tuple)):
            form = type(form.__name__, (RenderTemplateFormMixin, form), {})
            form.base_fields['render_template'].choices = choices
            kwargs['form'] = form
        return super().get_form(request, obj, **kwargs)

    def get_render_template(self, context, instance, placeholder):
        try:
            template = instance.glossary.get('render_template', self.get_template_choices()[0][0])
            get_template(template)  # check if template exists
        except (KeyError, IndexError, TemplateDoesNotExist, TypeError):
            template = self.render_template
        return template



import json

from django.conf.urls import url
from django.contrib.admin import site as default_admin_site
from django.contrib.admin.helpers import AdminForm
from django.core.exceptions import PermissionDenied
from django.forms import CharField, ModelChoiceField
from django.shortcuts import render
from django.template.response import TemplateResponse
from django.urls import reverse
from django.utils.http import urlencode
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _

from cms.plugin_base import CMSPluginBase, PluginMenuItem
from cms.plugin_pool import plugin_pool
from cms.toolbar.utils import get_plugin_tree_as_json
from cms.utils import get_language_from_request
from cmsplugin_cascade.clipboard.forms import ClipboardBaseForm
from cmsplugin_cascade.clipboard.utils import deserialize_to_clipboard, serialize_from_placeholder
from cmsplugin_cascade.models import CascadeClipboard


class CascadeClipboardPlugin(CMSPluginBase):
    render_plugin = False
    change_form_template = 'admin/cms/page/plugin/change_form.html'

    def get_plugin_urls(self):
        urlpatterns = [
            url(r'^export-plugins/$', self.export_plugins_view, name='export_clipboard_plugins'),
            url(r'^import-plugins/$', self.import_plugins_view, name='import_clipboard_plugins'),
        ]
        return urlpatterns

    @classmethod
    def get_extra_placeholder_menu_items(cls, request, placeholder):
        data = urlencode({
            'placeholder': placeholder.pk,
            'language': get_language_from_request(request),
        })
        return [
            PluginMenuItem(
                _("Export to Clipboard"),
                reverse('admin:export_clipboard_plugins') + '?' + data,
                data={},
                action='modal',
                attributes={
                    'icon': 'export',
                },
            ),
            PluginMenuItem(
                _("Import from Clipboard"),
                reverse('admin:import_clipboard_plugins') + '?' + data,
                data={},
                action='modal',
                attributes={
                    'icon': 'import',
                },
            )
        ]

    def render_modal_window(self, request, form):
        """
        Render a modal popup window with a select box to edit the form
        """
        opts = self.model._meta
        fieldsets = [(None, {'fields': list(form.fields)})]
        adminForm = AdminForm(form, fieldsets, {}, [])
        context = {
            **default_admin_site.each_context(request),
            'title': form.title,
            'adminform': adminForm,
            'add': False,
            'change': True,
            'save_as': False,
            'has_add_permission': False,
            'has_change_permission': True,
            'opts': opts,
            'root_path': reverse('admin:index'),
            'is_popup': True,
            'app_label': opts.app_label,
            'media': self.media + form.media,
        }
        return TemplateResponse(request, self.change_form_template, context)

    def import_plugins_view(self, request):
        # TODO: check for permissions

        title = _("Import from Clipboard")
        if request.method == 'GET':
            Form = type('ClipboardImportForm', (ClipboardBaseForm,), {
                'clipboard': ModelChoiceField(
                    queryset=CascadeClipboard.objects.all(),
                    label=_("Select Clipboard Content"),
                    required=False,
                ),
                'title': title,
            })
            form = Form(request.GET)
            assert form.is_valid()
        elif request.method == 'POST':
            Form = type('ClipboardImportForm', (ClipboardBaseForm,), {
                'clipboard': ModelChoiceField(
                    queryset=CascadeClipboard.objects.all(),
                    label=_("Select Clipboard Content"),
                ),
                'title': title,
            })
            form = Form(request.POST)
            if form.is_valid():
                return self.paste_from_clipboard(request, form)
        return self.render_modal_window(request, form)

    def paste_from_clipboard(self, request, form):
        placeholder = form.cleaned_data['placeholder']
        language = form.cleaned_data['language']
        cascade_clipboard = form.cleaned_data['clipboard']
        tree_order = placeholder.get_plugin_tree_order(language)
        deserialize_to_clipboard(request, cascade_clipboard.data)
        cascade_clipboard.last_accessed_at = now()
        cascade_clipboard.save(update_fields=['last_accessed_at'])

        # detach plugins from clipboard and reattach them to current placeholder
        cb_placeholder_plugin = request.toolbar.clipboard.cmsplugin_set.filter(plugin_type='PlaceholderPlugin').first()
        cb_placeholder_instance, _ = cb_placeholder_plugin.get_plugin_instance()
        new_plugins = cb_placeholder_instance.placeholder_ref.get_plugins()
        new_plugins.update(placeholder=placeholder)

        # reorder root plugins in placeholder
        root_plugins = placeholder.get_plugins(language).filter(parent__isnull=True).order_by('changed_date')
        for position, plugin in enumerate(root_plugins.iterator()):
            plugin.update(position=position)
        placeholder.mark_as_dirty(language, clear_cache=False)

        # create a list of pasted plugins to be added to the structure view
        all_plugins = placeholder.get_plugins(language)
        if all_plugins.exists():
            new_plugins = placeholder.get_plugins(language).exclude(pk__in=tree_order)
            data = json.loads(get_plugin_tree_as_json(request, list(new_plugins)))
            data['plugin_order'] = tree_order + ['__COPY__']
        else:
            return render(request, 'cascade/admin/clipboard_reload_page.html')
        data['target_placeholder_id'] = placeholder.pk
        context = {'structure_data': json.dumps(data)}
        return render(request, 'cascade/admin/clipboard_paste_plugins.html', context)

    def export_plugins_view(self, request):
        if not request.user.is_staff:
            raise PermissionDenied

        title = _("Export to Clipboard")
        if request.method == 'GET':
            Form = type('ClipboardExportForm', (ClipboardBaseForm,), {
                'identifier': CharField(required=False),
                'title': title,
            })
            form = Form(request.GET)
            assert form.is_valid()
        elif request.method == 'POST':
            Form = type('ClipboardExportForm', (ClipboardBaseForm,), {
                'identifier': CharField(),
                'title': title,
            })
            form = Form(request.POST)
            if form.is_valid():
                return self.add_to_clipboard(request, form)
        return self.render_modal_window(request, form)

    def add_to_clipboard(self, request, form):
        placeholder = form.cleaned_data['placeholder']
        language = form.cleaned_data['language']
        identifier = form.cleaned_data['identifier']
        data = serialize_from_placeholder(placeholder)
        CascadeClipboard.objects.create(
            identifier=identifier,
            data=data,
            created_by=request.user,
        )
        return render(request, 'cascade/admin/clipboard_close_frame.html', {})

plugin_pool.register_plugin(CascadeClipboardPlugin)



from django.forms.formsets import DELETION_FIELD_NAME
from django.forms.models import ModelForm

from entangled.forms import EntangledModelFormMixin, EntangledModelForm


class ManageChildrenFormMixin:
    """
    Classes derived from ``CascadePluginBase`` can optionally add this mixin class to their form,
    offering the input field ``num_children`` in their plugin editor. The content of this field is
    not persisted in the plugin's model.
    It allows the client to modify the number of children attached to this plugin.
    """
    class Meta:
        fields = ['num_children']

    def __init__(self, *args, **kwargs):
        instance = kwargs.get('instance')
        if instance:
            initial = {'num_children': instance.get_num_children()}
            kwargs.update(initial=initial)
        super().__init__(*args, **kwargs)


class CascadeModelFormMixin(EntangledModelFormMixin):
    """
    Classes inheriting from InlineAdmin and defining their own `form` shall use this special
    variant of an `EntangledModelForm`, otherwise deletion of inlined elements does not work.
    """
    def _clean_form(self):
        internal_fields = ['cascade_element', 'id', DELETION_FIELD_NAME]
        internal_cleaned_data = {key: val for key, val in self.cleaned_data.items() if key in internal_fields}
        super()._clean_form()
        self.cleaned_data.update(internal_cleaned_data)


class CascadeModelForm(CascadeModelFormMixin, ModelForm):
    """
    A convenience class to create a ModelForms to be used with djangocms-cascade
    """



from django.forms import widgets, CharField, ChoiceField
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _
from cms.plugin_pool import plugin_pool
from entangled.forms import EntangledModelFormMixin

from cmsplugin_cascade.plugin_base import CascadePluginBase


class HeadingFormMixin(EntangledModelFormMixin):
    TAG_TYPES = [('h{}'.format(k), _("Heading {}").format(k)) for k in range(1, 7)]

    tag_type = ChoiceField(
        choices=TAG_TYPES,
        label=_("HTML element tag"),
        help_text=_('Choose a tag type for this HTML element.')
    )

    content = CharField(
        label=_("Heading content"),
        widget=widgets.TextInput(attrs={'style': 'width: 100%; padding-right: 0; font-weight: bold; font-size: 125%;'}),
    )

    class Meta:
        entangled_fields = {'glossary': ['tag_type', 'content']}


class HeadingPlugin(CascadePluginBase):
    name = _("Heading")
    parent_classes = None
    allow_children = False
    form = HeadingFormMixin
    render_template = 'cascade/generic/heading.html'

    @classmethod
    def get_identifier(cls, instance):
        tag_type = instance.glossary.get('tag_type')
        content = mark_safe(instance.glossary.get('content', ''))
        if tag_type:
            return format_html('<code>{0}</code>: {1}', tag_type, content)
        return content

    def render(self, context, instance, placeholder):
        context = self.super(HeadingPlugin, self).render(context, instance, placeholder)
        context.update({'content': mark_safe(instance.glossary.get('content', ''))})
        return context

plugin_pool.register_plugin(HeadingPlugin)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.7 on 2016-06-17 07:52
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion

from cmsplugin_cascade import app_settings

plugins_with_sharables = app_settings.CMSPLUGIN_CASCADE['plugins_with_sharables'].keys()
plugins_with_sharables = ','.join(["'{}'".format(p) for p in plugins_with_sharables])


class Migration(migrations.Migration):

    dependencies = [
        ('cms', '0015_auto_20160421_0000'),
        ('cmsplugin_cascade', '0010_refactor_heading'),
    ]

    operations = [
        migrations.AddField(
            model_name='cascadeelement',
            name='shared_glossary',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='cmsplugin_cascade.SharedGlossary'),
        ),
        migrations.RunSQL(
            [("INSERT INTO cmsplugin_cascade_element (cmsplugin_ptr_id, glossary, shared_glossary_id) SELECT cmsplugin_ptr_id, glossary, shared_glossary_id FROM cmsplugin_cascade_sharableelement", None)],
            reverse_sql=[
                ("INSERT INTO cmsplugin_cascade_sharableelement (cmsplugin_ptr_id, glossary) SELECT cmsplugin_ptr_id, glossary FROM cmsplugin_cascade_element INNER JOIN cms_cmsplugin ON cmsplugin_ptr_id=cms_cmsplugin.id WHERE plugin_type IN ({})".format(plugins_with_sharables), None),
                ("DELETE FROM cmsplugin_cascade_element WHERE cmsplugin_ptr_id IN (SELECT cmsplugin_ptr_id FROM cmsplugin_cascade_sharableelement)", None)
            ]
        ),
        migrations.RemoveField(
            model_name='sharablecascadeelement',
            name='cmsplugin_ptr',
        ),
        migrations.RemoveField(
            model_name='sharablecascadeelement',
            name='shared_glossary',
        ),
        migrations.DeleteModel(
            name='SharableCascadeElement',
        ),
        migrations.CreateModel(
            name='SharableCascadeElement',
            fields=[
            ],
            options={
                'proxy': True,
            },
            bases=('cmsplugin_cascade.cascadeelement',),
        ),
    ]



from django.core.exceptions import ValidationError
from django.forms.fields import MultipleChoiceField
from django.forms.widgets import CheckboxSelectMultiple, MediaDefiningClass
from django.utils.translation import gettext_lazy as _

from entangled.forms import EntangledModelFormMixin


def remove_duplicates(lst):
    """
    Emulate what a Python ``set()`` does, but keeping the element's order.
    """
    dset = set()
    return [l for l in lst if l not in dset and not dset.add(l)]


def rectify_partial_form_field(base_field, partial_form_fields):
    """
    In base_field reset the attributes label and help_text, since they are overriden by the
    partial field. Additionally, from the list, or list of lists of partial_form_fields
    append the bound validator methods to the given base field.
    """
    base_field.label = ''
    base_field.help_text = ''
    for fieldset in partial_form_fields:
        if not isinstance(fieldset, (list, tuple)):
            fieldset = [fieldset]
        for field in fieldset:
            base_field.validators.append(field.run_validators)

def validate_link(link_data):
    """
    Check if the given model exists, otherwise raise a Validation error
    """
    from django.apps import apps

    try:
        Model = apps.get_model(link_data['model'])
        Model.objects.get(pk=link_data['pk'])
    except Model.DoesNotExist:
        raise ValidationError(_("Unable to link onto '{0}'.").format(Model.__name__))


def compute_aspect_ratio(image):
    if image.exif.get('Orientation', 1) > 4:
        # image is rotated by 90 degrees, while keeping width and height
        return float(image.width) / float(image.height)
    else:
        return float(image.height) / float(image.width)


def compute_aspect_ratio_with_glossary(glossary):
    if glossary['image']['exif_orientation'] > 4:
        # image is rotated by 90 degrees, while keeping width and height
        return float(glossary['image']['width']) / float(glossary['image']['height'])
    else:
        return float(glossary['image']['height']) / float(glossary['image']['width'])


def get_image_size(width, image_height, aspect_ratio):
    if image_height[0]:
        # height was given in px
        return (width, image_height[0])
    elif image_height[1]:
        # height was given in %
        return (width, int(round(width * image_height[1])))
    else:
        # as fallback, adopt height to current width
        return (width, int(round(width * aspect_ratio)))


def parse_responsive_length(responsive_length):
    """
    Takes a string containing a length definition in pixels or percent and parses it to obtain
    a computational length. It returns a tuple where the first element is the length in pixels and
    the second element is its length in percent divided by 100.
    Note that one of both returned elements is None.
    """
    responsive_length = responsive_length.strip()
    if responsive_length.endswith('px'):
        return (int(responsive_length.rstrip('px')), None)
    elif responsive_length.endswith('%'):
        return (None, float(responsive_length.rstrip('%')) / 100)
    return (None, None)


class CascadeUtilitiesMixin(metaclass=MediaDefiningClass):
    """
    If a Cascade plugin is listed in ``settings.CMSPLUGIN_CASCADE['plugins_with_extra_mixins']``,
    then this ``CascadeUtilitiesMixin`` class is added automatically to its plugin class in order to
    enrich it with utility classes, such as :class:`cmsplugin_cascade.bootstrap4.mixins.BootstrapUtilities`.
    """
    def __str__(self):
        return self.plugin_class.get_identifier(self)

    def get_form(self, request, obj=None, **kwargs):
        form = kwargs.get('form', self.form)
        assert issubclass(form, EntangledModelFormMixin), "Form must inherit from EntangledModelFormMixin"
        kwargs['form'] = type(form.__name__, (self.utility_form_mixin, form), {})
        return super().get_form(request, obj, **kwargs)

    @classmethod
    def get_css_classes(cls, obj):
        """Enrich list of CSS classes with customized ones"""
        css_classes = super().get_css_classes(obj)
        for utility_field_name in cls.utility_form_mixin.base_fields.keys():
            css_class = obj.glossary.get(utility_field_name)
            if css_class:
                if isinstance(css_class, str):
                    css_classes.append(css_class)
                elif isinstance(css_class, list):
                    css_classes.extend(css_class)
        return css_classes


class NamedCSSClasses:
    """
    Factory for building a class ``NamedCSSClassesMixin``. This class then is used as a mixin to
    all sorts of Cascade plugins. It shall be used to optionally activate preconfigured CSS classes
    for each configured plugin.
    To configure, pass in a list of 2-tuples, consisting of the CSS class plus a string describing
    this feature. For example:
    ```
    CMSPLUGIN_CASCADE['plugins_with_extra_mixins'] = {
        'LinkPlugin': NamedCSSClasses([
            ('stretched-link', "Use stretched Link"),
            ('full-width', "Full width"),
            …
        ]),
        …
    }
    ```

    The above configuration will add a checkbox for each entry in that list to the plugin editors.
    The administrator then can choose whether to add those classes to the HTML element or not.
    """
    def __new__(cls, choices):
        named_css_classes = MultipleChoiceField(
            label=_("Extra Styles"),
            choices=choices,
            widget=CheckboxSelectMultiple,
            required=False,
        )

        class Meta:
            entangled_fields = {'glossary': ['named_css_classes']}

        attrs = {
            'named_css_classes': named_css_classes,
            'Meta': Meta,
        }
        utility_form_mixin = type('UtilitiesFormMixin', (EntangledModelFormMixin,), attrs)
        return type('NamedCSSClassesMixin', (CascadeUtilitiesMixin,), {'utility_form_mixin': utility_form_mixin})



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations
from cmsplugin_cascade.models import CascadeElement


def forwards(apps, schema_editor):
    for cascade_element in CascadeElement.objects.all():
        if cascade_element.plugin_type != 'FramedIconPlugin':
            continue

        color = cascade_element.glossary.get('color')
        if isinstance(color, str):
            cascade_element.glossary['color'] = ('', color)
            cascade_element.save()
        shared = cascade_element.shared_glossary
        if shared:
            color = shared.glossary.get('color')
            if isinstance(color, str):
                shared.glossary['color'] = ('', color)
                shared.save()


def backwards(apps, schema_editor):
    for cascade_element in CascadeElement.objects.all():
        if cascade_element.plugin_type != 'FramedIconPlugin':
            continue

        color = cascade_element.glossary.get('color')
        if isinstance(color, (list, tuple)):
            cascade_element.glossary['color'] = color[1]
            cascade_element.save()
        shared = cascade_element.shared_glossary
        if shared:
            color = shared.glossary.get('color')
            if isinstance(color, (list, tuple)):
                shared.glossary['color'] = color[1]
                shared.save()


class Migration(migrations.Migration):

    dependencies = [
        ('cmsplugin_cascade', '0017_fake_proxy_models'),
    ]

    operations = [
        migrations.RunPython(forwards, reverse_code=backwards),
    ]



from django.forms import widgets
from django.forms.fields import BooleanField, CharField
from django.utils.translation import ngettext_lazy, gettext_lazy as _
from django.utils.text import Truncator
from django.utils.safestring import mark_safe
from django.forms.fields import IntegerField

from entangled.forms import EntangledModelFormMixin
from cms.plugin_pool import plugin_pool
from cmsplugin_cascade.forms import ManageChildrenFormMixin
from cmsplugin_cascade.plugin_base import TransparentWrapper, TransparentContainer
from cmsplugin_cascade.widgets import NumberInputWidget
from .plugin_base import BootstrapPluginBase


class TabSetFormMixin(ManageChildrenFormMixin, EntangledModelFormMixin):
    num_children = IntegerField(
        min_value=1,
        initial=1,
        widget=NumberInputWidget(attrs={'size': '3', 'style': 'width: 5em !important;'}),
        label=_("Tabs"),
        help_text=_("Number of tabs to start with."),
    )

    justified = BooleanField(
        label=_("Justified tabs"),
        required=False,
    )

    class Meta:
        untangled_fields = ['num_children']
        entangled_fields = {'glossary': ['justified']}


class BootstrapTabSetPlugin(TransparentWrapper, BootstrapPluginBase):
    name = _("Tab Set")
    parent_classes = ['BootstrapColumnPlugin']
    direct_child_classes = ['BootstrapTabPanePlugin']
    require_parent = True
    allow_children = True
    form = TabSetFormMixin
    render_template = 'cascade/bootstrap4/{}tabset.html'

    @classmethod
    def get_identifier(cls, instance):
        num_cols = instance.get_num_children()
        content = ngettext_lazy('with {} tab', 'with {} tabs', num_cols).format(num_cols)
        return mark_safe(content)

    def save_model(self, request, obj, form, change):
        wanted_children = int(form.cleaned_data.get('num_children'))
        super().save_model(request, obj, form, change)
        self.extend_children(obj, wanted_children, BootstrapTabPanePlugin)

plugin_pool.register_plugin(BootstrapTabSetPlugin)


class TabPaneFormMixin(EntangledModelFormMixin):
    tab_title = CharField(
        label=_("Tab Title"),
        widget=widgets.TextInput(attrs={'size': 80}),
    )

    class Meta:
        entangled_fields = {'glossary': ['tab_title']}


class BootstrapTabPanePlugin(TransparentContainer, BootstrapPluginBase):
    name = _("Tab Pane")
    direct_parent_classes = parent_classes = ['BootstrapTabSetPlugin']
    require_parent = True
    allow_children = True
    alien_child_classes = True
    form = TabPaneFormMixin

    @classmethod
    def get_identifier(cls, obj):
        content = obj.glossary.get('tab_title', '')
        if content:
            content = Truncator(content).words(3, truncate=' ...')
        return mark_safe(content)

plugin_pool.register_plugin(BootstrapTabPanePlugin)



#!/usr/bin/env python
from setuptools import setup, find_packages
from cmsplugin_cascade import __version__

with open('README.md', 'r') as fh:
    long_description = fh.read()

CLASSIFIERS = [
    'Development Status :: 5 - Production/Stable',
    'Environment :: Web Environment',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: MIT License',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Framework :: Django :: 2.1',
    'Framework :: Django :: 2.2',
    'Framework :: Django :: 3.0',
]

setup(
    name='djangocms-cascade',
    version=__version__,
    description='Build Single Page Applications using the Django-CMS plugin system',
    author='Jacob Rief',
    author_email='jacob.rief@gmail.com',
    url='https://github.com/jrief/djangocms-cascade',
    packages=find_packages(exclude=['examples', 'docs', 'tests']),
    install_requires=[
        'django>=2.1,<3.1',
        'django-classy-tags>=1.0',
        'django-cms>=3.5,<4',
        'django-entangled>=0.4',
        'djangocms-text-ckeditor>=3.7',
        'jsonfield',
        'requests',
    ],
    license='MIT',
    platforms=['OS Independent'],
    classifiers=CLASSIFIERS,
    long_description=long_description,
    long_description_content_type='text/markdown',
    include_package_data=True,
    zip_safe=False,
)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.10 on 2018-05-27 18:10
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('cmsplugin_cascade', '0018_iconfont_color'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='cascadeclipboard',
            options={'verbose_name': 'Persisted Clipboard Content', 'verbose_name_plural': 'Persisted Clipboard Contents'},
        ),
        migrations.AlterModelOptions(
            name='cascadeelement',
            options={'verbose_name': 'Element', 'verbose_name_plural': 'Elements'},
        ),
        migrations.AlterModelOptions(
            name='cascadepage',
            options={'verbose_name': 'Cascade Page Settings'},
        ),
        migrations.AlterModelOptions(
            name='segmentation',
            options={'managed': False, 'verbose_name': 'Segmentation'},
        ),
        migrations.AlterModelOptions(
            name='sortableinlinecascadeelement',
            options={'ordering': ['order']},
        ),
    ]



import logging
from django.core.exceptions import ValidationError
from django.forms import widgets, BooleanField, ChoiceField
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _
from entangled.forms import EntangledModelFormMixin

from cms.plugin_pool import plugin_pool
from cmsplugin_cascade import app_settings
from cmsplugin_cascade.fields import ColorField, MultiSizeField, CascadeImageField
from cmsplugin_cascade.image import ImagePropertyMixin
from cmsplugin_cascade.bootstrap4.plugin_base import BootstrapPluginBase
from cmsplugin_cascade.bootstrap4.container import ContainerGridMixin
from cmsplugin_cascade.bootstrap4.fields import BootstrapMultiSizeField
from cmsplugin_cascade.bootstrap4.picture import get_picture_elements

logger = logging.getLogger('cascade')


class ImageBackgroundMixin:
    @property
    def element_heights(self):
        element_heights = self.glossary.get('element_heights', {})
        for bp, media_query in self.glossary['media_queries'].items():
            if bp in element_heights:
                yield {'media': media_query['media'], 'height': element_heights[bp]}

    @property
    def background_color(self):
        try:
            color, disabled = self.glossary['background_color']
            if not disabled and disabled != 'disabled':
                return 'background-color: {};'.format(color)
        except (KeyError, TypeError, ValueError):
            pass
        return ''

    @property
    def background_attachment(self):
        try:
            return 'background-attachment: {background_attachment};'.format(**self.glossary)
        except KeyError:
            return ''

    @property
    def background_position(self):
        try:
            return 'background-position: {background_vertical_position} {background_horizontal_position};'.format(**self.glossary)
        except KeyError:
            return ''

    @property
    def background_repeat(self):
        try:
            return 'background-repeat: {background_repeat};'.format(**self.glossary)
        except KeyError:
            return ''

    @property
    def background_size(self):
        try:
            size = self.glossary['background_size']
            if size == 'width/height':
                size = self.glossary['background_width_height']
                return 'background-size: {width} {height};'.format(**size)
            else:
                return 'background-size: {};'.format(size)
        except KeyError:
            pass
        return ''


class JumbotronFormMixin(EntangledModelFormMixin):
    """
    Form class to validate the JumbotronPlugin.
    """
    ATTACHMENT_CHOICES = ['scroll', 'fixed', 'local']
    VERTICAL_POSITION_CHOICES = ['top', '10%', '20%', '30%', '40%', 'center', '60%', '70%', '80%', '90%', 'bottom']
    HORIZONTAL_POSITION_CHOICES = ['left', '10%', '20%', '30%', '40%', 'center', '60%', '70%', '80%', '90%', 'right']
    REPEAT_CHOICES = ['repeat', 'repeat-x', 'repeat-y', 'no-repeat']
    SIZE_CHOICES = ['auto', 'width/height', 'cover', 'contain']

    fluid = BooleanField(
        label=_("Is fluid"),
        initial=True,
        required=False,
        help_text=_("Shall this element occupy the entire horizontal space of its parent."),
    )

    element_heights = BootstrapMultiSizeField(
        label=("Element Heights"),
        required=True,
        allowed_units=['rem', 'px', 'auto'],
        initial='300px',
        help_text=_("This property specifies the height for each Bootstrap breakpoint."),
    )

    background_color = ColorField(
        label=_("Background color"),
    )

    image_file = CascadeImageField(
        label=_("Background image"),
        required=False,
    )

    background_repeat = ChoiceField(
        label=_("Background repeat"),
        choices=[(c, c) for c in REPEAT_CHOICES],
        widget=widgets.RadioSelect,
        initial='no-repeat',
        required=False,
        help_text=_("This property specifies how the background image repeates."),
    )

    background_attachment = ChoiceField(
        label=_("Background attachment"),
        choices=[(c, c) for c in ATTACHMENT_CHOICES],
        widget=widgets.RadioSelect,
        initial='local',
        required=False,
        help_text=_("This property specifies how to move the background image relative to the viewport."),
    )

    background_vertical_position = ChoiceField(
        label=_("Background vertical position"),
        choices=[(c, c) for c in VERTICAL_POSITION_CHOICES],
        initial='center',
        required=False,
        help_text=_("This property moves a background image vertically within its container."),
    )

    background_horizontal_position = ChoiceField(
        label=_("Background horizontal position"),
        choices=[(c, c) for c in HORIZONTAL_POSITION_CHOICES],
        initial='center',
        required=False,
        help_text=_("This property moves a background image horizontally within its container."),
    )

    background_size = ChoiceField(
        label=_("Background size"),
        choices=[(c, c) for c in SIZE_CHOICES],
        widget=widgets.RadioSelect,
        initial='auto',
        required=False,
        help_text=_("This property specifies how the background image is sized."),
    )

    background_width_height = MultiSizeField(
        ['width', 'height'],
        label=_("Background width/height"),
        allowed_units=['px', '%'],
        required=False,
        help_text=_("This property specifies the width and height of a background image in px or %."),
    )

    class Meta:
        entangled_fields = {'glossary': ['fluid', 'background_color', 'element_heights', 'image_file',
                                         'background_repeat', 'background_attachment',
                                         'background_vertical_position', 'background_horizontal_position',
                                         'background_size', 'background_width_height']}

    def validate_optional_field(self, name):
        field = self.fields[name]
        value = field.widget.value_from_datadict(self.data, self.files, self.add_prefix(name))
        if value in field.empty_values:
            self.add_error(name, ValidationError(field.error_messages['required'], code='required'))
        else:
            return value

    def clean(self):
        cleaned_data = super().clean()
        if cleaned_data['image_file']:
            self.validate_optional_field('background_repeat')
            self.validate_optional_field('background_attachment')
            self.validate_optional_field('background_vertical_position')
            self.validate_optional_field('background_horizontal_position')
            if self.validate_optional_field('background_size') == 'width/height':
                try:
                    cleaned_data['background_width_height']['width']
                except KeyError:
                    msg = _("You must at least set a background width.")
                    self.add_error('background_width_height', msg)
                    raise ValidationError(msg)
        return cleaned_data


class BootstrapJumbotronPlugin(BootstrapPluginBase):
    name = _("Jumbotron")
    model_mixins = (ContainerGridMixin, ImagePropertyMixin, ImageBackgroundMixin)
    require_parent = False
    parent_classes = ['BootstrapContainerPlugin', 'BootstrapColumnPlugin']
    allow_children = True
    alien_child_classes = True
    form = JumbotronFormMixin
    raw_id_fields = ['image_file']
    render_template = 'cascade/bootstrap4/jumbotron.html'
    ring_plugin = 'JumbotronPlugin'
    footnote_html = """<p>
    For more information about the Jumbotron please read the
    <a href="https://getbootstrap.com/docs/4.3/components/jumbotron/" target="_new">Bootstrap documentation</a>.
    </p>"""

    class Media:
        js = ['admin/js/jquery.init.js', 'cascade/js/admin/jumbotronplugin.js']

    def render(self, context, instance, placeholder):
        # image shall be rendered in a responsive context using the ``<picture>`` element
        try:
            elements = get_picture_elements(instance)
        except Exception as exc:
            logger.warning("Unable generate picture elements. Reason: {}".format(exc))
        else:
            try:
                if instance.child_plugin_instances and instance.child_plugin_instances[0].plugin_type == 'BootstrapRowPlugin':
                    padding='padding: {0}px {0}px;'.format(int(app_settings.CMSPLUGIN_CASCADE['bootstrap4']['gutter']/2))
                    context.update({'add_gutter_if_child_is_BootstrapRowPlugin': padding,})
                context.update({
                    'elements': [e for e in elements if 'media' in e] if elements else [],
                    'CSS_PREFIXES': app_settings.CSS_PREFIXES,
                })
            except Exception as exc:
                logger.warning("Unable generate picture elements. Reason: {}".format(exc))
        return self.super(BootstrapJumbotronPlugin, self).render(context, instance, placeholder)

    @classmethod
    def sanitize_model(cls, obj):
        sanitized = False
        super().sanitize_model(obj)
        grid_container = obj.get_bound_plugin().get_grid_instance()
        obj.glossary.setdefault('media_queries', {})
        for bp, bound in grid_container.bounds.items():
            obj.glossary['media_queries'].setdefault(bp.name, {})
            width = round(bound.max)
            if obj.glossary['media_queries'][bp.name].get('width') != width:
                obj.glossary['media_queries'][bp.name]['width'] = width
                sanitized = True
            if obj.glossary['media_queries'][bp.name].get('media') != bp.media_query:
                obj.glossary['media_queries'][bp.name]['media'] = bp.media_query
                sanitized = True
        return sanitized

    @classmethod
    def get_css_classes(cls, obj):
        css_classes = cls.super(BootstrapJumbotronPlugin, cls).get_css_classes(obj)
        if obj.glossary.get('fluid'):
            css_classes.append('jumbotron-fluid')
        else:
            css_classes.append('jumbotron')
        return css_classes

    @classmethod
    def get_identifier(cls, obj):
        identifier = super().get_identifier(obj)
        try:
            content = obj.image.name or obj.image.original_filename
        except AttributeError:
            content = _("Without background image")
        return format_html('{0}{1}', identifier, content)

plugin_pool.register_plugin(BootstrapJumbotronPlugin)



from django.forms.fields import ChoiceField
from django.utils.text import format_lazy
from django.utils.translation import gettext_lazy as _
from entangled.forms import EntangledModelFormMixin
from cmsplugin_cascade.utils import CascadeUtilitiesMixin
from cmsplugin_cascade.bootstrap4.grid import Breakpoint


class BootstrapUtilities:
    """
    Factory for building a class ``BootstrapUtilitiesMixin``. This class then is used as a mixin to
    all sorts of Bootstrap-4 plugins. Various Bootstrap-4 plugins are shipped using this mixin class
    in different configurations. These configurations can be overridden through the project's
    settings using:
    ```
    CMSPLUGIN_CASCADE['plugins_with_extra_mixins'] = {
        'Bootstrap<ANY>Plugin': BootstrapUtilities(
            BootstrapUtilities.background_and_color,
            BootstrapUtilities.margins,
            BootstrapUtilities.paddings,
            …
        ),
        …
    }
    ```

    The class ``BootstrapUtilities`` offers a bunch of property methods which return a list of
    input fields and/or select boxes. They then can be added to the plugin's editor. This is
    specially useful to add CSS classes from the utilities section of Bootstrap-4, such as
    margins, borders, colors, etc.
    """
    def __new__(cls, *args):
        form_fields = {}
        for arg in args:
            if isinstance(arg, property):
                form_fields.update(arg.fget(cls))

        class Meta:
            entangled_fields = {'glossary': list(form_fields.keys())}

        utility_form_mixin = type('UtilitiesFormMixin', (EntangledModelFormMixin,), dict(form_fields, Meta=Meta))
        return type('BootstrapUtilitiesMixin', (CascadeUtilitiesMixin,), {'utility_form_mixin': utility_form_mixin})

    @property
    def background_and_color(cls):
        choices = [
            ('', _("Default")),
            ('bg-primary text-white', _("Primary with white text")),
            ('bg-secondary text-white', _("Secondary with white text")),
            ('bg-success text-white', _("Success with white text")),
            ('bg-danger text-white', _("Danger with white text")),
            ('bg-warning text-white', _("Warning with white text")),
            ('bg-info text-white', _("Info with white text")),
            ('bg-light text-dark', _("Light with dark text")),
            ('bg-dark text-white', _("Dark with white text")),
            ('bg-white text-dark', _("White with dark text")),
            ('bg-transparent text-dark', _("Transparent with dark text")),
            ('bg-transparent text-white', _("Transparent with white text")),
        ]
        return {'background_and_color': ChoiceField(
            label=_("Background and color"),
            choices=choices,
            required=False,
            initial='',
        )}

    @property
    def margins(cls):
        form_fields = {}
        choices_format = [
            ('m-{}{}', _("4 sided margins ({})")),
            ('mx-{}{}', _("Horizontal margins ({})")),
            ('my-{}{}', _("Vertical margins ({})")),
            ('mt-{}{}', _("Top margin ({})")),
            ('mr-{}{}', _("Right margin ({})")),
            ('mb-{}{}', _("Bottom margin ({})")),
            ('ml-{}{}', _("Left margin ({})")),
        ]
        sizes = list(range(0, 6)) + ['auto']
        for bp in Breakpoint:
            if bp == Breakpoint.xs:
                choices = [(c.format('', s), format_lazy(l, s)) for c, l in choices_format for s in sizes]
                choices.insert(0, ('', _("No Margins")))
            else:
                choices = [(c.format(bp.name + '-', s), format_lazy(l, s)) for c, l in choices_format for s in sizes]
                choices.insert(0, ('', format_lazy(_("Inherit margin from {}"), previous_label)))
            previous_label = bp.label
            form_fields['margins_{}'.format(bp.name)] = ChoiceField(
                label=format_lazy(_("Margins for {breakpoint}"), breakpoint=bp.label),
                choices=choices,
                required=False,
                initial='',
            )
        return form_fields

    @property
    def vertical_margins(cls):
        form_fields = {}
        choices_format = [
            ('my-{}{}', _("Vertical margins ({})")),
            ('mt-{}{}', _("Top margin ({})")),
            ('mb-{}{}', _("Bottom margin ({})")),
        ]
        sizes = list(range(0, 6)) + ['auto']
        for bp in Breakpoint:
            if bp == Breakpoint.xs:
                choices = [(c.format('', s), format_lazy(l, s)) for c, l in choices_format for s in sizes]
                choices.insert(0, ('', _("No Margins")))
            else:
                choices = [(c.format(bp.name + '-', s), format_lazy(l, s)) for c, l in choices_format for s in sizes]
                choices.insert(0, ('', format_lazy(_("Inherit margin from {}"), previous_label)))
            previous_label = bp.label
            form_fields['margins_{}'.format(bp.name)] = ChoiceField(
                label=format_lazy(_("Margins for {breakpoint}"), breakpoint=bp.label),
                choices=choices,
                required=False,
                initial='',
            )
        return form_fields

    @property
    def paddings(cls):
        form_fields = {}
        choices_format = [
            ('p-{}{}', _("4 sided padding ({})")),
            ('px-{}{}', _("Horizontal padding ({})")),
            ('py-{}{}', _("Vertical padding ({})")),
            ('pt-{}{}', _("Top padding ({})")),
            ('pr-{}{}', _("Right padding ({})")),
            ('pb-{}{}', _("Bottom padding ({})")),
            ('pl-{}{}', _("Left padding ({})")),
        ]
        sizes = range(0, 6)
        for bp in Breakpoint:
            if bp == Breakpoint.xs:
                choices = [(c.format('', s), format_lazy(l, s)) for c, l in choices_format for s in sizes]
                choices.insert(0, ('', _("No Padding")))
            else:
                choices = [(c.format(bp.name + '-', s), format_lazy(l, s)) for c, l in choices_format for s in sizes]
                choices.insert(0, ('', format_lazy(_("Inherit padding from {}"), previous_label)))
            previous_label = bp.label
            form_fields['padding_{}'.format(bp.name)] = ChoiceField(
                label=format_lazy(_("Padding for {breakpoint}"), breakpoint=bp.label),
                choices=choices,
                required=False,
                initial='',
            )
        return form_fields

    @property
    def floats(cls):
        form_fields = {}
        choices_format = [
            ('float-{}none', _("Do not float")),
            ('float-{}left', _("Float left")),
            ('float-{}right', _("Float right")),
        ]
        for bp in Breakpoint:
            if bp == Breakpoint.xs:
                choices = [(c.format(''), l) for c, l in choices_format]
                choices.insert(0, ('', _("Unset")))
            else:
                choices = [(c.format(bp.name + '-'), l) for c, l in choices_format]
                choices.insert(0, ('', format_lazy(_("Inherit float from {}"), previous_label)))
            previous_label = bp.label
            form_fields['float_{}'.format(bp.name)] = ChoiceField(
                label=format_lazy(_("Floats for {breakpoint}"), breakpoint=bp.label),
                choices=choices,
                required=False,
                initial='',
            )
        return form_fields



import io
import mimetypes
import os
from django.conf import settings
from django.conf.urls import url
from django.core.exceptions import ViewDoesNotExist
from django.http.response import HttpResponse
from django.views.generic import TemplateView
from django.utils.cache import patch_cache_control
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _
from cms.app_base import CMSApp
from cms.apphook_pool import apphook_pool


class SphinxDocsView(TemplateView):
    def get(self, request, *args, **kwargs):
        slug = kwargs.get('slug', '')
        _, extension = os.path.splitext(slug)
        if extension in ['.png', '.jpg', '.jpeg', '.gif']:
            filename = os.path.join(settings.SPHINX_DOCS_ROOT, slug)
            content_type, _ = mimetypes.guess_type(filename)
            with io.open(filename, 'rb') as fd:
                response = HttpResponse(content=fd.read(), content_type=content_type)
                patch_cache_control(response, cache_control='max-age=86400')
                return response
        return super().get(request, page=slug, *args, **kwargs)

    def get_template_names(self):
        return [self.request.current_page.get_template()]

    def get_context_data(self, page='index.html', **kwargs):
        context = super().get_context_data(**kwargs)
        filename = os.path.join(settings.SPHINX_DOCS_ROOT, page, 'index.html')
        if not os.path.exists(filename):
            raise ViewDoesNotExist("{} does not exist".format(page))
        with io.open(filename, encoding='utf-8') as fd:
            context.update(page_content=mark_safe(fd.read()))
        return context


@apphook_pool.register
class SphinxDocsApp(CMSApp):
    name = _("Sphinx Documentation")

    def get_urls(self, page=None, language=None, **kwargs):
        return [
            url(r'^(?P<slug>\S+)/$', SphinxDocsView.as_view(), name='sphinx-documentation'),
        ]
